import numpy as np
import typing as tp

from abc import ABC


class TreeCoverageAnalyzer(ABC):
    def __init__(self):
        pass

class SkLearnTreeCoverageAnalyzer(TreeCoverageAnalyzer):
    """ This class acts as an analyzer for Scikit-Learn Trees (DecisionTreeClassifier/Regressor...)

    At the moment, it is only intended for evaluating the total leaf coverage achieved using a
    particular test data frame.

    Args:
        estimator: Fitted Scikit-Learn tree based classifier/regressor
    """
    def __init__(
        self,
        estimator: tp.Union['sklearn.tree.DecisionTreeClassifier',  # noqa: F821
                            'sklearn.tree.DecisionTreeRegressor']   # noqa: F821
    ):
        self.estimator = estimator

    def analyze(
            self,
            X: tp.Union['np.ndarray', 'pd.DataFrame', 'scipy.sparse.csr_matrix']  # noqa: F821
    ) -> "SkLearnTreeCoverageAnalyzer":
        """ Analyzes the tree. At the moment, only coverage is calculated.
        Analysis stats can be accesed after calling this method.

        Args:
            X: (test) Dataset which will be analyzed.

        Returns:
            self, the "fitted" Tree analyzer.
        """

        if hasattr(self, "all_leaves_"):
            del self.all_leaves_
            del self.covered_leaves_

        _ = self.get_all_leaves()
        coverage = self.get_covered_leaves(X)
        self.coverage_ = coverage
        return self

    def get_percent_coverage(self) -> float:
        """ Given test set (from the analyze method), get the percentage of total
        leaves that recieved at least one sample
        """
        num_unvisited = len(self.unvisited_leaves())
        return 1 - (num_unvisited / self.get_num_leaves())

    def get_all_leaves(self) -> np.array:
        """ Returns the node-id's of the leaves.
        """
        if not hasattr(self, "all_leaves_"):
            self.all_leaves_ = np.argwhere(self.estimator.tree_.children_right == -1).flatten()
        return self.all_leaves_

    def get_num_leaves(self) -> int:
        """ Returns the total number of leaves in the tree
        """
        return self.estimator.tree_.n_leaves

    def get_covered_leaves(
        self,
        X: tp.Union['np.ndarray', 'pd.DataFrame', 'scipy.sparse.csr_matrix']  # noqa: F821
    ) -> dict:
        """ Returns a dictionary in which the key is a node-id of a leaf and the value
        is the total number of samples of X that fell in each leaf.
        """
        if not hasattr(self, "covered_leaves_"):
            preds = self.estimator.apply(X)
            cnt = dict()
            for lid in self.get_all_leaves():
                cnt[lid] = np.sum(preds == lid)
            self.covered_leaves_ = cnt

        return self.covered_leaves_

    def unvisited_leaves(self):
        """ Returns a set with the node-ids of unvisited leafs.
        """
        if not hasattr(self, "unvisited_labels_leaves_"):
            self.unvisited_leaves_ = {key for key, val in self.coverage_.items() if val == 0}

        return self.unvisited_leaves_


class SkLearnTreeEnsembleCoverageAnalyzer(TreeCoverageAnalyzer):
    """ This class acts as an analyzer for Scikit-Learn Ensemble Tree based models
    (RandomForestClassifier/Regressor...) or anything that has an `estimators_` attribute
    where each estimator has a `tree_` attribute once fitted.

    At the moment, it is only intended for evaluating the total leaf coverage achieved using a
    particular test dataset.

    Args:
        estimator: Fitted Scikit-Learn tree based classifier/regressor
    """
    def __init__(
        self,
        estimator: tp.Union['sklearn.ensemble.RandomForestClassifier',  # noqa: F821
                            'sklearn.ensemble.RandomForestRegressor',   # noqa: F821
                            ]
    ):
        self.estimator = estimator
        self.tree_analyzers = [SkLearnTreeCoverageAnalyzer(est) for est in estimator.estimators_]

    def analyze(self, X) -> "SkLearnTreeEnsembleCoverageAnalyzer":
        """ Analyzes the ensemble. At the moment, only coverage is calculated.
        Analysis stats can be accesed after calling this method.

        Args:
            X: (test) Dataset which will be analyzed.

        Returns:
            self, the "fitted" Tree analyzer.
        """
        covs = [t.analyze(X) for t in self.tree_analyzers]
        self.coverage_ = covs
        return self

    def get_unvisited_leaves(self) -> tp.List[set]:
        """ Returns a list of sets in which each item i is the set of unvisited node-ids
        for each internal tree i.
        """
        unvisited_labels_ = list(
            filter(lambda a: len(a) > 0, [a.unvisited_leaves() for a in self.tree_analyzers])
        )
        return unvisited_labels_

    def get_covered_leaves(self) -> tp.List[dict]:
        """ Returns a list of dicts in which each item i is a dict counting how many items fell into one
        of the leaves of the tree i.
        """
        unvisited_labels_ = [a.covered_leaves_ for a in self.tree_analyzers]
        return unvisited_labels_

    def get_all_leaves(self) -> tp.List[np.array]:
        """ Returns a list of lists in which each item contains the node-ids of the leaves of
        each internal tree.
        """
        return [t.get_all_leaves() for t in self.tree_analyzers]

    def get_num_leaves(self) -> int:
        """ Returns the total number of leaves in the ensemble
        """
        return np.sum([t.get_num_leaves() for t in self.tree_analyzers])

    def get_percent_coverage(self) -> float:
        """ Given test set (from the analyze method), gets the percentage of total
        leaves that recieved at least one sample over all the internal trees in the ensemble.
        """
        total_leaves = np.sum(
            [a.get_num_leaves() for a in self.tree_analyzers]
        )

        num_unvisited = np.sum([len(a) for a in self.get_unvisited_leaves()])
        return 1 - (num_unvisited / total_leaves)


import numpy as np


def lin_combs_in_columns(X, tol=None):
    """
    Returns all the linear combinations that exists among the columns of X by performing Gauss-Jordan elimination.

    Args:
        X (np.array): Any numerical matrix (rank == 2) dimensions (r, c) that is either a numpy ndarray or can be converted into one
                      by passing it to the constructor of a numpy ndarray.
        tol (float):  An optional numerical tolerance. If not used, it will be set according to the number of columns squared times a small
                      constant. Note that numerical rounding errors will inevitably apply after performing reduction operations on the
                      matrix. In that case, deciding if a value is zero, or just very small depends on this tolerance. When you
                      independently compute the rank of X and expect a number of combinations, you can play with this tolerance to match
                      your expectations. Otherwise, it is okay to just leave the default value.

    Returns:
        None, if no linear combinations exist. If linear combinations exist: a matrix of shape (n, c) where n is the number of linear
        combinations found and c is the number of columns of the original X matrix. Each linear combination is defined so the
        **for each row** in the result matrix the product of X·r' is a column of all zeros. r' is the row transposed as a column vector.
    """
    X = np.array(list(X), dtype=np.float64)

    assert len(X.shape) == 2

    m = X.shape[1]

    if tol is None:
        tol = 1e6 * m * m * np.finfo(np.double).eps

    oidx = [i for i in range(m)]
    zX = 1 * (abs(X) > tol)
    cs = np.sum(zX, axis=0)
    clin = None

    for i, s in reversed(list(enumerate(cs))):
        if s == 0:
            cl = np.zeros((1, m), dtype=np.float64)
            cl[0, i] = 1
            if clin is None:
                clin = cl
            else:
                clin = np.vstack((clin, cl))

            if X.shape[1] == 1:
                return clin

            X = np.delete(X, i, axis=1)
            oidx.pop(i)

    rs = np.sum(zX, axis=1)

    for i, s in reversed(list(enumerate(rs))):
        if s == 0:
            X = np.delete(X, i, axis=0)

    nr, n = X.shape
    min_lc = max(0, n - nr)

    comb = np.zeros((n, m), dtype=np.float64)
    for j in range(n):
        comb[j, oidx[j]] = 1

    i = 0
    while i < n - 1 and i < nr:
        last = i == nr - 1
        row = X[i, :].copy()
        zrow = 1 * (abs(row) > tol)
        zpt = zrow[i]

        if zpt == 0:
            col = X[:, i].copy()
            zcol = 1 * (abs(col) > tol)

            j = i + 1
            while zcol[j] == 0:
                j += 1

            X[[i, j], :] = X[[j, i], :]

            row = X[i, :].copy()
            zrow = 1 * (abs(row) > tol)

        col = X[:, i].copy()
        j = i + 1

        while j < n:
            if zrow[j] > 0:
                f = -row[j] / row[i]

                X[:, j] = X[:, j] + f * col

                comb[j, :] = comb[j, :] + comb[i, :] * f

                if (last and n - j <= min_lc) or not np.any(abs(X[:, j]) > tol):
                    if clin is None:
                        clin = np.reshape(comb[j, :].copy(), (1, -1))
                    else:
                        clin = np.vstack((clin, comb[j, :]))

                    min_lc -= 1

                    comb = np.delete(comb, j, axis=0)
                    X = np.delete(X, j, axis=1)
                    row = np.delete(row, j)
                    zrow = np.delete(zrow, j)

                    n -= 1
                    j -= 1

            j += 1

        i += 1

    return clin


import pytest

from mercury.robust.basetests import RobustTest, TestState
from mercury.robust.errors import FailedTestError

class DummyTest(RobustTest):

    def __init__(self, test_param="succeed", name=None, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.test_param = test_param

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

        if self.test_param == "fail":
            raise FailedTestError("test has failed")
        elif self.test_param == "exception":
            raise Exception("test has raised exception")

def test_robust_test():

    # Test that succeeds
    dummy_test = DummyTest(name="test1")
    assert dummy_test.name == "test1"
    assert dummy_test._state == TestState.NOT_EXECUTED
    dummy_test.run_safe()
    assert dummy_test._state == TestState.SUCCESS

    # Test that fails
    dummy_test = DummyTest(test_param="fail")
    assert dummy_test._state == TestState.NOT_EXECUTED
    dummy_test.run_safe()
    assert dummy_test._state == TestState.FAIL

    # Test that raises exception
    dummy_test = DummyTest(test_param="exception")
    assert dummy_test._state == TestState.NOT_EXECUTED
    dummy_test.run_safe()
    assert dummy_test._state == TestState.EXCEPTION



import numpy as np
import pandas as pd
import sklearn
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.metrics import accuracy_score, r2_score, mean_absolute_error
import inspect
import copy
from typing import Union, Callable, List, Dict
from mercury.robust.basetests import RobustModelTest, TaskInferrer
from mercury.robust.errors import FailedTestError

from mercury.dataschema.feature import FeatType
from mercury.dataschema import DataSchema
from mercury.monitoring.drift.drift_simulation import BatchDriftGenerator
from ._tree_analysis import SkLearnTreeCoverageAnalyzer, SkLearnTreeEnsembleCoverageAnalyzer



class ModelReproducibilityTest(RobustModelTest):
    """
    This test checks if the training of a model is reproducible. It does so by training the model two times
    and checking whether they give the same evaluation metric and predictions. If the difference in
    the evaluation metric is higher than `threshold_eval` parameter then the test fails. Similarly, if
    the percentage of different predictions is higher than `threshold_yhat` then the test fails.
    You can check only one of the checks (the evaluation metric and predictions) or only one of them.

    Args:
        model (BaseEstimator, tf.keras.Model): Unfitted model that we are checking reproducibility
        train_dataset (pd.DataFrame): The pandas dataset used for training the model.
        target (str): The name of the target variable predicted by the model which must be one column in
            the train dataset.
        train_fn (Callable): function called to train the model. The interface of the function is
            train_fn(model, X, y, train_params) and returns the fitted model.
        train_params (dict): Params to use for training. It is passed as a parameter to the `train_fn`
        eval_fn (Callable): function called to evaluate the model. The interface of the function is
            eval_fn(model, X, y, eval_params) and returns a float. If None, then the check of looking if
            training two times produces the same evaluation metric won't be performed.
        eval_params (dict): Params to use for evaluation. It is passed as a parameter to the `eval_fn`.
        threshold_eval (float): difference that we are able to tolerate in the evaluation function in order
            to pass the test. If the difference of the evaluation metric when training the model two times
            is higher than the threshold, then the test fails. Default value is 0
        predict_fn (Callable): function called to get the predictions of a dataset once the model is trained.
            The interface of the function is predict_fn(model, X, predict_params) and returns the predictions.
        predict_params (dict): Params to use for prediction. It is passed as a parameter to the `predict_fn`
        threshold_yhat (float): If `predict_fn` is given, this is the percentage of different predictions that we are
            can tolerate without making the test fail. A prediction from the model trained two times is considered
            different according to the parameter `yhat_allowed_diff`. Default value for `threshold_yhat` is 0,
            meaning that with just one prediction different the test will fail.
        yhat_allowed_diff (float): difference that we can tolerate in order to consider that a sample has the
            same prediction from two models. If a prediction of the model trained two times differ by more than
            `yhat_allowed_diff` then that prediction is considered to be different. Default value is 0.
        test_dataset (pd.DataFrame):  If given, a separate dataset with identical column structure used for the
            evaluation parts.
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> # Define necessary functions for model training, evaluation, and getting predictions
        >>> def train_model(model, X, y, train_params=None):
        >>>     model.fit(X, y)
        >>>     return model
        >>> def eval_model(model, X, y, eval_params=None):
        >>>     y_pred = model.predict(X)
        >>>     return accuracy_score(y, y_pred)
        >>> def get_predictions(model, X, pred_params=None):
        >>>     return model.predict(X)
        >>> # Create and run test
        >>> from mercury.robust.model_tests import ModelReproducibilityTest
        >>> test = ModelReproducibilityTest(
        >>>     model = model,
        >>>     train_dataset = df_train,
        >>>     target = "label_col",
        >>>     train_fn = train_model_fn,
        >>>     eval_fn = eval_model_fn,
        >>>     threshold_eval = 0,
        >>>     predict_fn = get_predictions_fn,
        >>>     epsilon_yhat = 0,
        >>>     test_dataset = df_test
        >>> )
        >>> test.run()
        ```
    """

    def __init__(
        self, model: Union["BaseEstimator", "tf.keras.Model"],  # noqa: F821
        train_dataset: "pd.DataFrame",  # noqa: F821,
        target: str,
        train_fn: Callable, train_params: dict = None,
        eval_fn: Callable = None, eval_params: dict = None, threshold_eval: float = 0.,
        predict_fn: Callable = None, predict_params: dict = None,
        threshold_yhat: float = 0., yhat_allowed_diff: float = 0.,
        test_dataset: "pd.DataFrame" = None,  # noqa: F821,
        name: str = None,
        *args, **kwargs
    ):
        super().__init__(model, name, *args, **kwargs)
        self.train_dataset = train_dataset
        self.train_fn = train_fn
        self.train_params = train_params
        self.target = target
        self.eval_fn = eval_fn
        self.eval_params = eval_params
        self.threshold_eval = threshold_eval
        self.predict_fn = predict_fn
        self.predict_params = predict_params
        self.threshold_yhat = threshold_yhat
        self.yhat_allowed_diff = yhat_allowed_diff
        self.test_dataset = test_dataset

    def _check_diff_in_eval(self, model_1, model_2, X, y):
        self.eval_1 = self.eval_fn(model_1, X, y, self.eval_params)
        self.eval_2 = self.eval_fn(model_2, X, y, self.eval_params)
        if abs(self.eval_1 - self.eval_2) > self.threshold_eval:
            return True
        return False

    def _check_diff_in_predictions(self, model_1, model_2, X):
        yhat_1 = self.predict_fn(model_1, X, self.predict_params)
        yhat_2 = self.predict_fn(model_2, X, self.predict_params)
        self.diff = (np.abs(yhat_1 - yhat_2) > self.yhat_allowed_diff).mean()
        if self.diff > self.threshold_yhat:
            return True
        return False

    def _clone_unfitted_model(self, model):
        """
        Returns a new copy of the model
        """
        if isinstance(model, sklearn.base.BaseEstimator):
            return sklearn.base.clone(model)
        else:
            # import tensorflow only if we need it
            import tensorflow as tf
            if isinstance(model, (tf.keras.Model, tf.estimator.Estimator)):
                return tf.keras.models.clone_model(model)
            else:
                # Check if we have deepcopy method
                members, _ = zip(*inspect.getmembers(model))
                if '__deepcopy__' in members:
                    return copy.deepcopy(model)
                # Otherwise, we cannot copy the model
                raise ValueError("The type of model used is not compatible with the test")

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if training is not reproducible
        """

        super().run(*args, **kwargs)

        X_train = self.train_dataset.loc[:, self.train_dataset.columns != self.target]
        y_train = self.train_dataset.loc[:, self.target]
        if self.test_dataset is not None:
            X_test = self.test_dataset.loc[:, self.test_dataset.columns != self.target]
            y_test = self.test_dataset.loc[:, self.target]

        # Clone the model
        model_1 = self._clone_unfitted_model(self.model)
        model_2 = self._clone_unfitted_model(self.model)

        # Train the models
        if self.train_fn is None:
            raise ValueError("You must provide a valid train_fn to train the model")
        model_1 = self.train_fn(model_1, X_train, y_train, self.train_params)
        model_2 = self.train_fn(model_2, X_train, y_train, self.train_params)

        # Check that at least one of eval_fn or predict_fn are specified
        if self.eval_fn is None and self.predict_fn is None:
            raise ValueError("At least one of eval_fn or predict_fn must be specified")

        if self.eval_fn is not None:
            # Check evaluation on train
            if self._check_diff_in_eval(model_1, model_2, X_train, y_train):
                raise FailedTestError(
                    f"Eval metric different in train dataset when training two times ({round(self.eval_1,4)} vs {round(self.eval_2,4)}). "
                    f"The max difference allowed is {round(self.threshold_eval, 4)} "
                    f"The model is not reproducible."
                )
            # Check evaluation on test
            if (self.test_dataset is not None) and self._check_diff_in_eval(model_1, model_2, X_test, y_test):
                raise FailedTestError(
                    f"Eval metric different in test dataset when training two times ({round(self.eval_1,4)} vs {round(self.eval_2,4)}). "
                    f"The max difference allowed is {round(self.threshold_eval, 4)} "
                    f"The model is not reproducible."
                )

        if self.predict_fn is not None:
            # Check evaluation on train
            if self._check_diff_in_predictions(model_1, model_2, X_train):
                raise FailedTestError(
                    f"Percentage of different predictions in training set is {round(self.diff,4)} when training two times and the "
                    f"maximum allowed is {round(self.threshold_yhat,4)} "
                    f"The model is not reproducible."
                )
            # Check evaluation on test
            if (self.test_dataset is not None) and self._check_diff_in_predictions(model_1, model_2, X_test):
                raise FailedTestError(
                    f"Percentage of different predictions in test set is {round(self.diff,4)} when training two times and the "
                    f"maximum allowed is {round(self.threshold_yhat,4)} "
                    f"The model is not reproducible."
                )


class ModelSimplicityChecker(RobustModelTest, TaskInferrer):

    """
    This test looks if a trained model has a simple baseline which trained in the same dataset
    gives better or similar performance on a test dataset. If not specified, the baseline is considered
    a LogisticRegression model for classification tasks and LinearRegression model for regression tasks.

    Args:
        model (Union["BaseEstimator", "tf.keras.Model"]): the trained model which we will compare against
            a baseline model
        X_train (Union["pandas.DataFrame", np.ndarray]): features of train dataset used to train the `model`.
            This same dataset will be used to train the baseline.
        y_train (Union["pandas.DataFrame", np.ndarray]): targets of train dataset used to train the `model`.
            This same dataset will be used to train the baseline.
        X_test (Union["pandas.DataFrame", np.ndarray]): features of test dataset which will be used to evaluate
            the `model` and the baseline.
        y_test (Union["pandas.DataFrame", np.ndarray]): targets of test dataset which will be used to evaluate
            the `model` and the baseline.
        ignore_feats: Features which won't be used in the `baseline_model`. Only use when `X_train` and `X_test`
            are pandas dataframes.
        baseline_model ("BaseEstimator"): Optional model that will be used as a baseline. It doesn't have to be
            an sklearn model, however, it needs to implement the `fit()` and `predict()` methods. If not specified,
            a LogisticRegression is used in case of classification and LinearRegression in case of regression.
        task (str): Task of the dataset. It must be either 'classification' or 'regression'. If None provided
              then it will be auto inferred from the `target` column.
        eval_fn (Callable[[np.ndarray], np.ndarray]): function which returns a metric to compare the performance of
            the `model` against the baseline. The interface of the function is eval_fn(y_true, y_pred). Note that in
            this test is assumed that the higher the metric the better the model, therefore if you use a metric which
            lower means better then the `eval_fn` should return the negative of the metric. If not specified, the
            accuracy score will be used in case of classification and R2 in case of regression.
        predict_fn (Callable[["BaseEstimator"], np.ndarray]): Custom predict function to obtain predictions from `model`.
            The interface of the function is predict_fn(model, X_test). Note that by default this is None, and in that case,
            the test will try to obtain the predictions using the `predict()` method of the `model`. That works in many cases
            where you are using scikit-learn models and the predict function returns what the `eval_fn` is expecting.
            However, in some case you might need to define a custom `predict_fn`. For example, when using a tf.keras
            model and the accuracy_score as eval_fn, the `predict()` method from tf.keras model returns the
            probabilities and the accuracy_score expects the classes, therefore you can use the `predict_fn` to obtain
            the classes. Another alternative is to pass the already computed predictions using the `test_predictions`
            parameter.
        test_predictions (np.array): array of predictions of the test set obtained by the `model`. If given, the
            test will use them instead of computing them using the `predict()` method of the `model` or the `predict_fn`
            This might be useful in cases where you are creating multiple tests and you want to avoid to compute
            the predictions each time.
        threshold (float): The threshold to use when comparing the `model` and the baseline. It is used to establish
            the limit to consider that the baseline performs similar or better than the `model`. Concretely, if the
            baseline model performs worse than the `model` with a difference equal or higher than this threshold
            then the test passes. Otherwise, if the baseline model performs better or performs worse but with a
            difference lower than this threshold, then the test fails.
        name (str): A name for the test. If not used, it will take the name of the class.
        encode_cat_feats (bool): bool to indicate whether to encode categorical features as
            one-hot-encoding. Note that if you specify it as False and your dataset has string column then the test will
            raise an exception since the default baseline models won't be able to deal with string columns. Default
            value is True
        scale_num_feats (bool): bool to indicate whether to scale the numeric features. If True, a StandardScaler
            is used. Default value is True.
        schema_custom_feature_map: Internally, this test generates a DataSchema object. In case you find it makes
                                   wrong feature type assignations to your features you can pass here a dictionary which
                                   specify the feature type of the columns you want to fix. (See DataSchema. `force_types`
                                   parameter for more info on this).
        dataset_schema: Pre built schema. This argument is complementary to `schema_custom_feature_map`. In case you want to
                        manually build your own DataSchema object, you can pass it here and the test will internally use it
                        instead of the default, automatically built. If you provide this parameter, `schema_custom_feature_map`
                        will not be used.

    Example:
        ```python
        >>> from mercury.robust.model_tests import ModelSimplicityChecker
        >>> test = ModelSimplicityChecker(
        >>>     model = model,
        >>>     X_train = X_train,
        >>>     y_train = y_train,
        >>>     X_test = X_test,
        >>>     y_test = y_test,
        >>>     threshold = 0.02,
        >>>     eval_fn = roc_auc_score
        >>> )
        >>> test.run()
        ```
    """

    def __init__(
        self, model: Union["BaseEstimator", "tf.keras.Model"],  # noqa: F821
        X_train: Union["pandas.DataFrame", np.ndarray],  # noqa: F821
        y_train: Union["pandas.Series", np.ndarray],  # noqa: F821
        X_test: Union["pandas.DataFrame", np.ndarray],  # noqa: F821
        y_test: Union["pandas.Series", np.ndarray],  # noqa: F821
        baseline_model: "BaseEstimator" = None,  # noqa: F821
        ignore_feats: List[str] = None,
        task: str = None,
        eval_fn: Callable[[np.ndarray], np.ndarray] = None,
        predict_fn: Callable[["BaseEstimator"], np.ndarray] = None,  # noqa: F821
        threshold: float = None,
        name: str = None,
        encode_cat_feats: bool = True,
        scale_num_feats: bool = True,
        test_predictions: np.array = None,
        schema_custom_feature_map: Dict[str, FeatType] = None,
        dataset_schema: DataSchema = None,
        *args, **kwargs
    ):

        super().__init__(model, name, *args, **kwargs)

        self.X_train = X_train
        self.y_train = y_train
        self.X_test = X_test
        self.y_test = y_test
        self.ignore_feats = ignore_feats if ignore_feats is not None else []

        self.task = task
        self.baseline_model = baseline_model

        self.threshold = threshold
        self.eval_fn = eval_fn
        self.predict_fn = predict_fn
        self.test_predictions = test_predictions

        self.encode_cat_feats = encode_cat_feats
        self.scale_num_feats = scale_num_feats

        self._schema_custom_feature_map = schema_custom_feature_map
        self._dataset_schema = dataset_schema

        self.metric_model = None
        self.metric_baseline_model = None

    def _get_pipeline_with_preprocessing(
        self,
        num_feats: List[str],
        cat_feats: List[str],
        model: "BaseEstimator"  # noqa: F821
    ):

        l_transformers = []
        if (self.scale_num_feats) and (len(num_feats) > 0):
            l_transformers.append(("num", StandardScaler(), num_feats))
        if (self.encode_cat_feats) and (len(cat_feats) > 0):
            l_transformers.append(("cat", OneHotEncoder(handle_unknown="ignore"), cat_feats))

        if len(l_transformers) > 0:
            preprocessor = ColumnTransformer(transformers=l_transformers, remainder='drop')
            return Pipeline(steps=[("preprocessor", preprocessor), ("classifier", model)])
        else:
            return Pipeline(steps=[("classifier", model)])

    def _get_baseline_model(self, schema):
        """
        Obtain eval_fn and model_baseline, depending on the task
        """
        if self.task == 'classification':
            baseline_model = LogisticRegression() if self.baseline_model is None else self.baseline_model
        else:
            baseline_model = LinearRegression() if self.baseline_model is None else self.baseline_model

        # Create pipeline with feature preprocessing
        cat_feats = schema.categorical_feats + schema.binary_feats
        cat_feats = [f for f in cat_feats if f not in self.ignore_feats]
        num_feats = schema.continuous_feats + schema.discrete_feats
        num_feats = [f for f in num_feats if f not in self.ignore_feats]

        baseline_model = self._get_pipeline_with_preprocessing(num_feats, cat_feats, baseline_model)

        return baseline_model

    def info(self):
        if self.metric_model is not None:
            return {"metric_model": self.metric_model, "metric_baseline_model": self.metric_baseline_model}
        else:
            return None

    def run(self, *args, **kwargs):

        X_train = self.X_train
        y_train = self.y_train
        X_test = self.X_test
        y_test = self.y_test

        # Obtain Task if not specified
        if not self.task:
            self.task = self.infer_task(y_train)

        if self.task not in ('classification', 'regression'):
            raise ValueError("Error. 'task' must be either classification or regression")

        # Generate Schema (if Xs are not pandas, transform them first)
        if not isinstance(X_train, pd.DataFrame):
            X_train = pd.DataFrame(X_train)
        if not isinstance(X_test, pd.DataFrame):
            X_test = pd.DataFrame(X_test)

        if self._dataset_schema is None:
            schema_train = DataSchema().generate(X_train, force_types=self._schema_custom_feature_map)
        else:
            schema_train = self._dataset_schema

        # Obtain model_baseline, depending on the task
        baseline_model = self._get_baseline_model(schema_train)
        baseline_name = str(baseline_model[-1])

        # Obtain eval function, depending on the task
        if self.task == 'classification':
            eval_fn = accuracy_score if self.eval_fn is None else self.eval_fn
        else:
            eval_fn = r2_score if self.eval_fn is None else self.eval_fn

        # Train the baseline model
        baseline_model = baseline_model.fit(X_train, y_train)

        # Evaluate on test set
        if self.test_predictions is not None:
            self.metric_model = eval_fn(y_test, self.test_predictions)
        elif self.predict_fn is not None:
            self.metric_model = eval_fn(y_test, self.predict_fn(self.model, X_test))
        else:
            self.metric_model = eval_fn(y_test, self.model.predict(X_test))
        self.metric_baseline_model = eval_fn(y_test, baseline_model.predict(X_test))

        # Compare difference
        if (self.metric_model - self.threshold) < self.metric_baseline_model:
            raise FailedTestError(
                f"""a {baseline_name} baseline model gives an evaluation metric of {round(self.metric_baseline_model, 4)} while """
                f"""the original model gives an evaluation metric of {round(self.metric_model, 4)}. The difference is lower than """
                f"""the threshold {round(self.threshold, 4)}"""
            )


class DriftPredictionsResistanceTest(RobustModelTest):
    """
    This test checks the robustness of a trained model to drift in the X dataset. It uses the model to predict the Y from the given X and
    uses that `Y` as a ground truth. Then, it applies some drift to the data in X by using a `BatchDriftGenerator` object and does a new
    prediction `drifted_Y` using the drifted dataset. If both the `Y` and `drifted_Y` diverge by more that some given tolerance value,
    the test fails. This test does only one verification. If we need doing more than one drift check, just apply multiple tests with
    appropriate names to simplify following up the results.

    Args:
        model: The model being evaluated. The model must be already trained and will not be trained again by this test. It is assumed to
            have a sklearn-like compliant predict() method that works on the dataset and returns a vector that is accepted by the
            evaluation function.
        X: A pandas dataset that can be used by the model's predict() method and whose predicted values will be used as the ground truth
            drift measurement.
        drift_type: The name of the method of a BatchDriftGenerator specifying the type of drift to be applied. E.g., "shift_drift",
            "scale_drift", ... You can check the class BatchDriftGenerator in _drift_simulation to see all available types
        drift_args: A dictionary with the argument expected by the drift method. E.g., {cols: ['a', 'b'], iqr: [1.12, 1.18]} for
            "scale_drift".
        names_categorical: An optional list with the names of the categorical variables. If this is used, the internal `BatchDriftGenerator`
            will use a `DataSchema` object to fully define the variables in X as either categorical (if in the list) or continuous
            (otherwise). This allows automatically selecting the columns without using the `cols` argument in `drift_args`. If this
            parameter is not given, the `DataSchema` is not initally defined and either you select the columns manually by declaring a
            `cols` argument in `drift_args` or the `BatchDriftGenerator` will create a `DataSchema` than automatically infers the
            column types.
        dataset_schema: Alternatively, you can provide a pre built schema for an even higher level of control. If you use this argument,
            `names_categorical` is not used. The schema fully defines binary, categorical, discrete or continuous. If you still define
            the `cols` argument in `drift_args`, that selection will prevail over whatever is in `dataset_schema`.
        eval: If given, an evaluation function that defines how "different" the predictions are. The function must accept two vectors
            returned by model.predict() and return some positive value that indicates the difference in the predictions and is compared
            with `tolerance`. If not given, then a sum of squared differences will be used unless the `model.predict()` method generates
            the hard labels for a multiclass classification problem. In this last case, the `eval` function will be a function to compute
            the number of different predictions.
        tolerance: A real value to be compared with the result of the evaluation function. Note that the purpose of the test is to check
            if the model is robust to the introduced drift. Therefore, the test will fail when the result (named as `loss`) is higher than
            the tolerance, meaning the model predictions considerably change with the introduced drift. When the test fails, you can see the
            value returned by the `eval` function in the RuntimeError message displayed as `loss`.
        name: A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.model_tests import DriftPredictionsResistanceTest
        >>> test = DriftPredictionsResistanceTest(
        >>>     model = trained_model,
        >>>     X = X,
        >>>     drift_type = "shift_drift",
        >>>     drift_args = {'cols': ['feature_1'], 'force': 100.},
        >>>     tolerance = 5,
        >>> )
        >>> test.run()
        ```
    """

    def __init__(
        self,
        model,
        X,
        drift_type,
        drift_args,
        names_categorical=None,
        dataset_schema=None,
        eval=None,
        tolerance=1e-3,
        name=None,
    ):
        super().__init__(model, name)

        self.model = model
        self.X = X.copy()
        self.Y = model.predict(X)

        if dataset_schema is not None:
            self.gen = BatchDriftGenerator(X=self.X, schema=dataset_schema)
        elif names_categorical is not None:
            self.gen = BatchDriftGenerator(
                X=self.X,
                schema=DataSchema().generate_manual(
                    dataframe=self.X,
                    categ_columns=names_categorical,
                    discrete_columns=[],
                    binary_columns=[],
                ),
            )
        else:
            self.gen = BatchDriftGenerator(X=self.X)

        self.fun = getattr(self.gen, drift_type, None)

        if not callable(self.fun):
            raise RuntimeError(
                "drift_type = %s must be a method of BatchDriftGenerator."
            )

        self.drift_args = drift_args
        self.eval = eval
        self.tolerance = tolerance
        self.loss = None

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            `FailedTestError` with a descriptive message if any of the attempts fail.
        """

        super().run(*args, **kwargs)

        drifted_Y = self._calculate_drifted_Y()

        if self.eval is None:
            eval_fn = self._get_default_eval_fn()
            self.loss = eval_fn(self.Y, drifted_Y)
        else:
            self.loss = self.eval(self.Y, drifted_Y)

        if self.loss > self.tolerance:
            raise FailedTestError(
                "Test failed. Prediction loss drifted above tolerance = %.3f (loss = %.3f)"
                % (self.tolerance, self.loss)
            )

    def info(self) -> dict:
        return {'loss': self.loss}

    def _calculate_drifted_Y(self):
        drifted_gen = self.fun(**self.drift_args)
        drifted_Y = self.model.predict(drifted_gen.data)
        return drifted_Y

    def _get_default_eval_fn(self):

        if self._contains_hard_labels(self.Y):
            return lambda y, drifted_y: np.sum(y != drifted_y)
        else:
            return lambda y, drifted_y: np.sum((y - drifted_y) ** 2)

    def _contains_hard_labels(self, y):
        return np.all(np.mod(y, 1)) == 0 and np.any(y > 1)


class DriftMetricResistanceTest(DriftPredictionsResistanceTest, TaskInferrer):
    """
    This test checks the robustness of a trained model to drift in the X dataset. It uses the model to predict the Y_no_drifted
    from the given X. Then, it applies some drift to the data in X by using a `BatchDriftGenerator` object and calculates
    Y_drifted. Then calculates a metric using Y_true with Y_no_drifted on the one hand and using Y_true with Y_drifted on the other
    hand. If the difference between these two metrics diverge more than some given tolerance value, the test fails.
    This test does only one verification. If we need doing more than one drift check, just apply multiple tests with
    appropriate names to simplify following up the results.

    Args:
        model: The model being evaluated. The model must be already trained and will not be trained again by this test. It is assumed to
            have a sklearn-like compliant predict() method that works on the dataset and returns a vector that is accepted by the
            evaluation function.
        X: A pandas dataset that can be used by the model's predict() method and whose predicted values will be used as the ground truth
            drift measurement.
        Y: array with the ground truth values. It will be used to calculate the metric for the non-drifted dataset and for the drifted
            dataset
        drift_type: The name of the method of a BatchDriftGenerator specifying the type of drift to be applied. E.g., "shift_drift",
            "scale_drift", ... You can check the class BatchDriftGenerator in _drift_simulation to see all available types
        drift_args: A dictionary with the argument expected by the drift method. E.g., {cols: ['a', 'b'], iqr: [1.12, 1.18]} for
            "scale_drift".
        names_categorical: An optional list with the names of the categorical variables. If this is used, the internal `BatchDriftGenerator`
            will use a `DataSchema` object to fully define the variables in X as either categorical (if in the list) or continuous
            (otherwise). This allows automatically selecting the columns without using the `cols` argument in `drift_args`. If this
            parameter is not given, the `DataSchema` is not initally defined and either you select the columns manually by declaring a
            `cols` argument in `drift_args` or the `BatchDriftGenerator` will create a `DataSchema` than automatically infers the
            column types.
        dataset_schema: Alternatively, you can provide a pre built schema for an even higher level of control. If you use this argument,
            `names_categorical` is not used. The schema fully defines binary, categorical, discrete or continuous. If you still define
            the `cols` argument in `drift_args`, that selection will prevail over whatever is in `dataset_schema`.
        eval: the evaluation function to use to calculate the metric. If passed, the interface of the function must be
            `eval_fn(y_true, y_hat)`. If not used, the mean absolute error will be used for regression and the accuracy for
            classification.
        tolerance: A real value to be compared with the difference of the computed metric with the non-drifted dataset and with the
            dirfted dataset.
        task: 'classification' or 'regression'. If not given, the test will try to infer it from `Y`
        name: A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> testing_dataset = pd.DataFrame(...)
        >>> rf = RandomForestClassifier().fit(train_data)
        >>> drift_args = {'cols': 'feature_1', 'method': 'percentile', 'method_params': {'percentile': 95}}
        >>> test = DriftMetricResistanceTest(
        ...    model = rf,
        ...    X = testing_dataset[features],
        ...    Y = testing_dataset[target],
        ...    drift_type = 'outliers_drift',
        ...    drift_args = drift_args,
        ...    tolerance = 0.05
        ... )
        >>> test.run()  # The test will fail if the difference in the metric is more than 0.05
        ```
    """

    def __init__(
        self,
        model,
        X: pd.DataFrame,
        Y: np.array,
        drift_type,
        drift_args,
        names_categorical=None,
        dataset_schema=None,
        eval=None,
        tolerance=None,
        task=None,
        name=None
    ):
        super().__init__(model, X, drift_type, drift_args, names_categorical, dataset_schema, eval, tolerance, name)
        self.Y_true = Y
        self.task = task
        self.metric_no_drifted = None
        self.metric_drifted = None
        self.metric_diff = None

    def run(self, *args, **kwargs):

        drifted_Y = self._calculate_drifted_Y()
        self.metric_diff = self._evaluate_performance_degradation(drifted_Y)
        if self.metric_diff > self.tolerance:
            raise FailedTestError(
                "Test failed. The metric of drifted dataset has changed above tolerance = %.3f (diff = %.3f)"
                % (self.tolerance, self.metric_diff)
            )

    def info(self) -> dict:
        return {
            'metric_no_drifted': self.metric_no_drifted,
            'metric_drifted': self.metric_drifted,
            'metric_diff': self.metric_diff,
        }

    def _evaluate_performance_degradation(self, drifted_Y):

        if self.eval is None:
            eval_fn = self._get_default_eval_fn()
        else:
            eval_fn = self.eval

        self.metric_no_drifted = eval_fn(self.Y_true, self.Y)
        self.metric_drifted = eval_fn(self.Y_true, drifted_Y)

        return np.abs(self.metric_no_drifted - self.metric_drifted)

    def _get_default_eval_fn(self):
        task = self.task if self.task is not None else self.infer_task(self.Y_true)
        if task == "classification":
            return accuracy_score
        elif task == "regression":
            return mean_absolute_error
        else:
            raise ValueError("Invalid task. It must be 'classification' or 'regression'")


class ClassificationInvarianceTest(RobustModelTest):
    """
    The idea of the `ClassificationInvarianceTest` is to check that if we apply a label-preserving perturbation the prediction of the
    model shouldn't change.

    This class helps to check this by checking the number of samples where the conditional of preserving the label doesn't hold and raising
    an error if the percentage of samples where the label is not preserved is higher than a specified threshold. We must pass to the test
    the original samples and the already generated perturbed samples.

    When calling run(), a exception `FailedTestError` is raised if the test fails. Additonally, the next attributes are filled:

        - preds_original_samples: stores the predictions for the original samples
        - preds_perturbed_samples: stores the predictions for perturbed samples
        - pred_is_different: stores for each sample a boolean array indicating if the predictions for the perturbed samples are different
            to the original sample
        - num_failed_per_sample: stores for each sample the number of perturbations where the prediction is different to the original sample
        - num_perturbed_per_sample: stores for each samples the number of perturbations
        - samples_with_errors: boolean array containing which samples contain errors
        - rate_samples_with_errors: the percentage of samples that contains at least one pertubed sample that the model predicted
            a different label.
        - total_rate_errors: the total percentage of perturbed samples that the model predicted a different label

    This test is based on the paper:
        ['Beyond Accuracy: Behavioral Testing of NLP Models with CheckList'](https://homes.cs.washington.edu/~marcotcr/acl20_checklist.pdf)

    Args:
        original_samples (Union(List[str], np.array)): List or array containing the original samples
        perturbed_samples (Union(List[List[str]], np.array)): List or array containing the perturbed samples. Each element
            of the list or each row of the array contains one or several perturbed samples corresponding to the sample in the
            same position/index in `original_samples`
        model: The model being evaluated. The model must be already trained. It is assumed to have a sklearn-like compliant
            predict() method that works on the `original_samples` and `perturbed_samples` and returns a vector with the
            the predictions. Alternatively, you can pass a `predict_fn`
        predict_fn (Callable): function that given the samples returns the predicted labels. Only used if `model` argument is
            None.
        threshold (float): if the percentage of samples with errors is higher than this threshold, then a `FailedTestError`
            will be raised. Default value is 0.05
        check_total_errors_rate (bool): this indicates what to consider as percentage of errors. If True, then each perturbed
            sample counts to calculate the rate. If False, then the rate is calculated with the number of samples indepedently
            of how many perturbations each sample has. Default value is True
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> original_samples = ["sample1", "sample2"]
        >>> perturbed_samples = [
        ...    "pertubed_sample_1 for sample1", "perturbed_sample_2 for sample1",
        ...    "perturbed_sample_1 for sample2", "perturbed_sample2 for sample2"
        ... ]
        >>> test = ClassificationInvarianceTest(
        ...    original_samples,
        ...    pertubed_samples,
        ...    predict_fn=my_model.predict,
        ...    threshold=0.1,
        ...    check_total_errors_rate=True,
        ...    name="Invariance Test"
        ... )
        >>> test.run()
        ```
    """

    def __init__(
        self,
        original_samples: Union[List[str], np.array],
        perturbed_samples: Union[List[List[str]], np.array],
        model: "BaseEstimator" = None,  # noqa: F821
        predict_fn: Callable = None,
        threshold: float = 0.05,
        check_total_errors_rate: bool = True,
        name: str = None,
        *args, **kwargs
    ):
        super().__init__(model, name, *args, **kwargs)
        self.original_samples = original_samples
        self.perturbed_samples = perturbed_samples
        self.predict_fn = predict_fn
        self.threshold = threshold
        self.check_total_errors_rate = check_total_errors_rate
        self.num_samples = len(self.original_samples)

        # Attributes for results
        self.preds_original_samples = None
        self.preds_perturbed_samples = None
        self.pred_is_different = None
        self.num_failed_per_sample = None
        self.num_perturbed_per_sample = None
        self.samples_with_errors = None
        self.rate_samples_with_errors = None
        self.total_rate_errors = None

    def run(self, *args, **kwargs):
        """run the test"""

        self._check_args()
        self.preds_original_samples = self._get_predictions(self.original_samples)
        self._check_changes_in_perturbed_samples_predictions()
        self._calculate_errors_rate()
        self._fail_test_if_high_error()

    def info(self):
        if self.preds_original_samples is not None:
            return {
                "rate_samples_with_errors": self.rate_samples_with_errors,
                "total_rate_errors": self.total_rate_errors
            }
        else:
            return None

    def get_examples_failed(self, n_samples: int = 5, n_perturbed: int = 1):
        """
        Returns examples of samples that failed.

        Args:
            - n_samples (int): number of samples to recover.
            - n_perturbed (int): for each sample, how many failed perturbations to recover.
        """

        selected_failed_samples = self._select_random_failed_samples(n_samples)

        # Get perturbations that failed for each selected sample
        examples = []
        for idx_selected in selected_failed_samples:
            selected_perturbed = self._select_random_failed_perturbations(idx_selected, n_perturbed)
            for idx_perturbed in selected_perturbed:
                sample_original = self.original_samples[idx_selected]
                sample_perturbed = self.perturbed_samples[idx_selected][idx_perturbed]
                pred_original = self.preds_original_samples[idx_selected]
                pred_perturbed = self.preds_perturbed_samples[idx_selected][idx_perturbed]
                examples.append((sample_original, sample_perturbed, pred_original, pred_perturbed))

        return pd.DataFrame(examples, columns=["original", "perturbed", "pred_original", "pred_perturbed"])

    def _check_args(self):
        if len(self.original_samples) != len(self.perturbed_samples):
            raise ValueError("Arguments original_samples and perturbed_samples must have the same length")

        if self.model is None and self.predict_fn is None:
            raise ValueError("At least one of the arguments 'model' or 'predict_fn must be specified")

    def _get_predictions(self, samples):
        if self.model is not None:
            return self.model.predict(samples)
        else:
            return self.predict_fn(samples)

    def _check_changes_in_perturbed_samples_predictions(self):
        """For each one of the samples, check in which perturbed samples the prediction changes"""
        self.preds_perturbed_samples = []
        self.pred_is_different = []
        self.num_failed_per_sample = []
        self.num_perturbed_per_sample = []
        for i in range(self.num_samples):
            # Note if improvement in speed is need: flatten perturbed samples, and make the predictions in bigger batches
            pred_is_different = self._check_changes_in_predictions(self.preds_original_samples[i], self.perturbed_samples[i])
            self.pred_is_different.append(pred_is_different)
            self.num_failed_per_sample.append(np.sum(pred_is_different))
            self.num_perturbed_per_sample.append(len(self.perturbed_samples[i]))

    def _check_changes_in_predictions(self, original_pred, perturbed_samples):
        """For a particular sample, checks in which perturbed samples the predictions changes"""
        self.preds_perturbed_samples.append(self._get_predictions(perturbed_samples))
        pred_is_different = np.array(self.preds_perturbed_samples[-1]) != original_pred
        return pred_is_different

    def _calculate_errors_rate(self):
        self.samples_with_errors = np.array(self.num_failed_per_sample) > 0
        self.rate_samples_with_errors = self.samples_with_errors.sum() / len(self.original_samples)
        self.total_rate_errors = np.sum(self.num_failed_per_sample) / np.sum(self.num_perturbed_per_sample)

    def _fail_test_if_high_error(self):
        """raises a `FailedTestError` if the error is higher than the threshold"""
        if self.check_total_errors_rate:
            error_rate = self.total_rate_errors
        else:
            error_rate = self.rate_samples_with_errors
        if error_rate > self.threshold:
            raise FailedTestError(f"Error rate {error_rate} is higher than threshold {self.threshold}")

    def _select_random_failed_samples(self, n_samples):
        """Randomly, select samples that failed"""
        idx_failed_samples = np.where(np.array(self.num_failed_per_sample) > 0)[0]
        selected_failed_samples = np.random.choice(idx_failed_samples, size=min(n_samples, len(idx_failed_samples)), replace=False)
        return selected_failed_samples

    def _select_random_failed_perturbations(self, idx_sample, n_perturbed):
        """Randomly, select perturbations that failed of a specific sample"""
        preds_is_diff = np.where(self.pred_is_different[idx_sample])[0]
        selected_perturbed = np.random.choice(preds_is_diff, size=min(n_perturbed, len(preds_is_diff)), replace=False)
        return selected_perturbed


class TreeCoverageTest(RobustModelTest):
    """
    This test checks whether a given test_dataset covers a minimum percentage of all the branches of
    a tree. Use this in case you want to make sure no leaves are left unexplored when testing your model.
    In case the percentage of coverage is less than the required `threshold_coverage`, the test will fail.

    Right now, this test only supports scikit-learn tree models, including sklearn pipelines with one tree
    model in one of its steps.

    TODO: Add support for other frameworks such as lightgbm, catboost or xgboost.

    Args:
        model: Fitted tree-based model (or sklearn pipeline with a tree-based model) to inspect
        test_dataset: Dataset for testing the coverage.
        threshold_coverage: this threshold represents the minimum percentage that the `test_dataset` needs
            to cover in order to pass the test. Eg. if `threshold_coverage=0.7` then the `test_dataset`
            needs to cover at least 70% of the branches to pass the test.
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> testing_dataset = pd.DataFrame(...)
        >>> rf = RandomForestClassifier().fit(train_data)
        >>> test = TreeCoverageTest(
        ...    rf,
        ...    testing_dataset,
        ...    threshold_coverage=.8
        ...    name="My Tree Coverage Test"
        ... )
        >>> test.run()  # The test will fail if the obtained coverage is less than 80%
        ```
    """

    def __init__(
        self, model: Union["DecisionTreeClassifier", "DecisionTreeRegressor",
                           "RandomForestClassifier", "RandomForestRegressor",
                           "sklearn.pipeline.Pipeline"],  # noqa: F821
        test_dataset: "pd.DataFrame",  # noqa: F821,
        threshold_coverage: float = .7,
        name: str = None,
        *args, **kwargs
    ):
        super().__init__(model, name, *args, **kwargs)
        self.test_dataset = test_dataset
        self.threshold_coverage = threshold_coverage
        if isinstance(model, sklearn.pipeline.Pipeline):
            index_model = self._identify_tree_in_pipeline(model)
            self.test_dataset = self._execute_pipeline_until_step(model, index_model, self.test_dataset)
            model = model.steps[index_model][1]
        self._analyzer = self._get_analyzer(model)
        self.coverage = None

    def _get_analyzer(self, model):
        """
        Builds an analyzer according to the model type.
        """
        if isinstance(model, (RandomForestRegressor, RandomForestClassifier)):
            return SkLearnTreeEnsembleCoverageAnalyzer(model)
        elif isinstance(model, (DecisionTreeClassifier, DecisionTreeRegressor)):
            return SkLearnTreeCoverageAnalyzer(model)
        else:
            raise ValueError("Error. This model type is not supported by this test.")

    def _identify_tree_in_pipeline(self, pipeline):
        supported_tree_types = (
            RandomForestRegressor,
            RandomForestClassifier,
            DecisionTreeClassifier,
            DecisionTreeRegressor
        )
        index_model = -1
        for i, stage in enumerate(pipeline.steps):
            if isinstance(stage[1], supported_tree_types):
                index_model = i
        if index_model == -1:
            raise ValueError("A pipeline without any supported model types by this test has been provided")
        return index_model

    def _execute_pipeline_until_step(self, pipeline, index_step, data):
        return pipeline[:index_step].transform(data)

    def run(self, *args, **kwargs):
        """Run the test"""
        self._analyzer.analyze(self.test_dataset)
        coverage = self._analyzer.get_percent_coverage()
        self.coverage = coverage

        if coverage < self.threshold_coverage:
            raise FailedTestError(f"Achieved a coverage of {coverage} while the minimum required was {self.threshold_coverage}")

    def info(self) -> dict:
        return {'coverage': self.coverage}
    
class FeatureCheckerTest(RobustModelTest):
    """
    This model robustness test checks if training the models using less columns in the dataframe can achieve identical results.
    To do so, it uses the variable importance taken from the model itself or estimated using a mercury.explainability explainer
    (ShuffleImportanceExplainer). It does a small number of attempts at removing unimportant variables and "fails if it succeeds",
    since success implies that a smaller, therefore more efficient, dataset should be used instead. The purpose of this test is
    not to find that optimal dataset. That can be achieved by removing the columns identified as unimportant and iterating.

    **NOTE**: This class will retrain (fit) the model several times resulting in the model being altered as a side effect. Make
    copies of your model before using this tool. This tool is intended as a diagnostic tool.

    Args:
        model: The model being evaluated. The model is assumed to comply to a minimalistic sklearn-like interface. More precisely:
            1. It must have a fit() method that works on the dataset and the dataset with some columns removed. It is important that each
            time the method fit() is called the model is trained from scratch (ie does not perform incremental training).
            2. It must have a predict() method that works on the dataset and returns a vector that is accepted by the evaluation function.
            3. If the argument `importance` is used, it must have an attribute with that name containing a list of (value, column_name)
            tuples which is consistent with many sklearn models.
            Alternatively, you can provide a function that creates a model or pipeline. This is useful in models or pipelines where
            removing columns from a dataframe raises an error because it expects the removed column. In this case, the interface of
            the function is model_fn(dataframe, model_args), where dataframe is the input pandas dataframe with the features already
            removed and model_fn_args is a dictionary with optional parameters that can be passed to the function. Importantly,
            this function just needs to create the model (unfitted) instance, but not to perform the training
        model_fn_args: if you are using a function in `model` parameter, you can use this argument to provide arguments to that function.
        train: The pandas dataset used for training the model, possibly with some columns removed.
        target: The name of the target variable predicted by the model which must be one columns in the train dataset.
        test: If given, a separate dataset with identical column structure used for the evaluation parts. Otherwise, the train dataset
            will be used instead.
        importance: If given, the name of a property in the model is updated by a fit() call. It must contain the importance of the
            columns as a list of (value, column_name) tuples. Otherwise, the importance of the variables will be estimated using a
            mercury.explainer ShuffleImportanceExplainer.
        eval: If given, an evaluation function that defines what "identical" results are. The function must accept two vectors returned
            by model.predict() and return some positive value that is smaller than `tolerance` if "identical". Otherwise a sum of squared
            differences will be used instead.
        tolerance: A real value to be compared with the result of the evaluation function. Note that the purpose of the test is finding
            unimportant variables. Therefore, the test will fail when the result (named as `loss`) is smaller than the tolerance, meaning
            the model could work "identically" well with less variables. When the test fails, you can see the value returned by the `eval`
            function in the RuntimeError message displayed as `loss`.
        num_tries: The total number of column removal tries the test should do before passing. This value times `remove_num` must be
            smaller than the number of columns (Y excluded).
        remove_num: The number of columns removed at each try.
        name: A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.model_tests import FeatureCheckerTest
        >>> test = FeatureCheckerTest(
        >>>     model=model,
        >>>     train=df_train,
        >>>     target="label_col",
        >>>     test=df_test,
        >>>     num_tries=len(df_train.columns)-1,
        >>>     remove_num=1,
        >>>     tolerance=len(df_test)*0.01
        >>> )
        >>> test.run()
        ```
    """

    def __init__(
        self,
        model: Union["Estimator", Callable],  # noqa: F821
        train: pd.DataFrame,
        target: str,
        test: pd.DataFrame = None,
        model_fn_args: dict = None,
        importance: str = None,
        eval: Callable = None,
        tolerance: float = 1e-3,
        num_tries: int = 3,
        remove_num: int = 1,
        name: str = None
    ):
        super().__init__(model, name)
        self.model_fn_args = model_fn_args
        self.train = train
        self.target = target
        self.test = test if test is not None else train
        self.importance = importance
        self.eval = eval
        self.remove_num = remove_num
        self.num_tries = num_tries
        self.tolerance = tolerance
        self.losses = {}

    def info(self):
        return self.losses

    def _get_least_important(self, N):

        names = []

        if self.importance is None:

            try:
                from mercury.explainability.explainers import ShuffleImportanceExplainer
            except:
                raise ImportError(
                    'Install mercury-explainability to use the FeatureCheckerTest with this type of model'
                )

            def model_inference_function(data, target):
                assert target == self.target

                X = data.loc[:, data.columns != self.target]
                Y = data.loc[:, self.target]
                Yh = self.fitted_model.predict(X)

                if self.eval is None:
                    return sum((Y - Yh) ** 2)

                return self.eval(Y, Yh)

            explainer = ShuffleImportanceExplainer(model_inference_function)

            explanation = explainer.explain(self.test, target=self.target)

            for name, _ in sorted(explanation.get_importances(), key=lambda x: x[1]):
                names.append(name)
                N -= 1
                if N == 0:
                    break

            return names

        try:
            importance = getattr(self.model, self.importance)

        except AttributeError:
            return names

        finally:
            columns = self.test.columns[self.test.columns != self.target]
            if len(columns) != len(importance):
                return names

            for _, name in sorted(zip(importance, columns)):
                names.append(name)
                N -= 1
                if N == 0:
                    break

            return names

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            `FailedTestError` with a descriptive message if any of the attempts fail.
        """

        super().run(*args, **kwargs)

        N = self.remove_num * self.num_tries

        X_train = self.train.loc[:, self.train.columns != self.target]
        Y_train = self.train.loc[:, self.target]

        # Fit model with all the features
        self._fit_model(X_train, Y_train)

        X_test = self.test.loc[:, self.test.columns != self.target]
        Y_hat = self.fitted_model.predict(X_test)

        remove = self._get_least_important(N)

        if len(remove) != N:
            raise RuntimeError(
                "Wrong arguments. Not enough columns for remove_num*num_tries."
            )

        for t in range(self.num_tries):
            exclude = remove[0:self.remove_num]
            remove = remove[self.remove_num:]

            exclude.append(self.target)

            x_cols = [c for c in self.train.columns if c not in exclude]

            X_alt_train = self.train.loc[:, x_cols]

            # Fit model with the removed features
            self._fit_model(X_alt_train, Y_train)

            X_alt_test = self.test.loc[:, x_cols]
            Y_alt_hat = self.fitted_model.predict(X_alt_test)

            if self.eval is None:
                loss = sum((Y_hat - Y_alt_hat) ** 2)
            else:
                loss = self.eval(Y_hat, Y_alt_hat)
            self.losses[", ".join(exclude[:-1])] = loss

            if loss < self.tolerance:
                raise FailedTestError(
                    "Test failed. A model fitted removing columns [%s] is identical within tolerance %.3f (loss = %.3f)"
                    % (", ".join(exclude[:-1]), self.tolerance, loss)
                )

    def _fit_model(self, X, Y):
        if callable(self.model):
            self.fitted_model = self.model(X, self.model_fn_args)
            self.fitted_model.fit(X, Y)
        else:
            self.fitted_model = self.model.fit(X, Y)



from .basetests import RobustTest

import pandas as pd
from typing import List, Union


class TestSuite():
    """
    A TestSuite runs a list of tests. It has two modes of running them. By setting the `run_safe=True`
    all the tests are run and the results of each test are available at the end of the execution using
    the method `get_results_as_df()`. If we set `run_safe=False` a `FailedTestError` exception will
    be raised if a test fails.

    Args:
        tests (List[RobustTest]): A list of RobustTest objects that will be executed
        run_safe (bool): If True, all the tests will be executed independently if some of them fail or
            not. If False, a FailedTestError will be raised when the first test fails and the rest
            won't be executed

    Example:
        ```python
        >>> # Create tests
        >>> drift_test = DriftTest(df_test, train_schma, name="drift_train_test")
        >>> lin_comb_test = LinearCombinationsTest(df_train)
        >>> # Create the TestSuite with the tests
        >>> test_suite = TestSuite(
        >>>     tests=[schema_test, drift_test, lin_comb_test, label_leak_test, noisy_label_test], run_safe=True
        >>> )
        >>> # Obtain results
        >>> test_results = test_suite.get_results_as_df()
        ```
    """

    def __init__(self, tests: List[RobustTest], run_safe: bool = True, *args, **kwargs):
        self._tests = tests
        self._run_safe = run_safe

    def run(self, tests_to_run: List[Union[str, int]] = None):
        """
        Executes all the tests. If attribute `run_safe` is True, an exception is raised when the first
        tests fails. If attribute `run_safe` is False, it executes all the tests even if some of them
        fail.

        Args:
            tests_to_run: If we don't want to run all the test, we can provide this argument with the list
                of index or names of the tests what we want to run.
        """
        for i, test in enumerate(self._tests):
            if self._test_should_run(test, i, tests_to_run):
                if self._run_safe:
                    test.run_safe()
                else:
                    test.run()

    def _test_should_run(self, test, i, tests_to_execute = None):

        if tests_to_execute is None:
            return True
        elif (test.name in tests_to_execute) or (i in tests_to_execute):
            return True
        else:
            return False

    def get_results_as_df(self):
        """
        Obtains a dataframe with a summary of the tests and their execution states

        Returns:
            Pandas DataFrame with the test results

        """
        df_results = pd.DataFrame(
            data={
                "name": [test.name for test in self._tests],
                "state": [test.get_state() for test in self._tests],
                "error": [test.get_error_message() for test in self._tests],
                "info": [test.info() for test in self._tests],
            }
        )
        return df_results


from abc import ABC
from enum import Enum
import pandas as pd

from mercury.robust.errors import FailedTestError

from mercury.dataschema.feature import (
    FeatureFactory,
    ContinuousFeature,
    DiscreteFeature,
    CategoricalFeature,
    BinaryFeature
)


class TestState(Enum):
    """
    Enumeration class with the different states that a Test can be.
    """
    SUCCESS = 'SUCCESS'
    FAIL = 'FAIL'
    NOT_EXECUTED = 'NOT_EXECUTED'
    EXCEPTION = 'EXCEPTION'


class RobustTest(ABC):
    """
    Base class for the robust testing implementations. It keeps three attributes:

        1) `_name` (str): name of the test. It can be specified when constructing the test, otherwise
            it will take the name of the class.
        2) `_state `(TestState): state in which the test is found. Initially, NOT_EXECUTED
        3) `_error_message` (str): After running the test, this contains an error message if the
            test fails or runs an exception

    """
    def __init__(self, name=None, *args, **kwargs):
        self._name = name
        self._state = TestState.NOT_EXECUTED
        self._error_mesage = ""

    @property
    def name(self) -> str:
        if self._name is None:
            return self.__class__.__name__
        else:
            return self._name

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    def info(self) -> dict:
        return None

    def get_error_message(self) -> str:
        return self._error_mesage

    def get_state(self) -> TestState:
        return self._state

    def run(self, *args, **kwargs):
        """
        Run the test, raising FailedTestError exception if it fails
        It must be implemented in subclasses
        """
        pass

    def run_safe(self, *args, **kwargs):
        """
        Runs the test without raising FailedTestError exception. Instead, the attribute
        `_state` will store the state that has reached the test, and the attribute
        `_error_message` will contain a message in case that an error happens.
        """
        try:
            self.run(*args, **kwargs)
            self._state = TestState.SUCCESS
            self._error_mesage = ""
        except FailedTestError as e:
            self._state = TestState.FAIL
            self._error_mesage = str(e)
        except Exception as e:
            self._state = TestState.EXCEPTION
            self._error_mesage = e.__repr__()


class RobustDataTest(RobustTest):
    def __init__(self, base_dataset, name=None, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.base_dataset = base_dataset


class RobustModelTest(RobustTest):
    def __init__(self, model, name=None, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.model = model


class TaskInferrer():
    def infer_task(
        self,
        y: "np.array"  # noqa: F821
    ) -> str:

        # Case multiclass classification with labels encoded as vectors
        if (len(y.shape) == 2) and (y.shape[1] > 1):
            return 'classification'

        feature_factory = FeatureFactory()
        y_feature = feature_factory.build_feature(pd.Series(y))
        if isinstance(y_feature, BinaryFeature) or isinstance(y_feature, CategoricalFeature):
            return 'classification'
        elif isinstance(y_feature, ContinuousFeature) or isinstance(y_feature, DiscreteFeature):
            return 'regression'
        else:
            raise ValueError(
                "Unable to infer task from 'y'. Provide a valid 'y' or specify "
                "the task with 'task' param "
            )


from sklearn.tree import (
    DecisionTreeRegressor,
    DecisionTreeClassifier
)
from sklearn.metrics import mean_squared_error, accuracy_score
from sklearn.base import (
    BaseEstimator,
    TransformerMixin
)
from sklearn.utils.validation import check_is_fitted


import warnings
import numpy as np


class TreeSelector(TransformerMixin, BaseEstimator):
    """
    Feature selection using shallow trees.
    This selector will order the features by its predictive power w.r.t. the
    target variable. The purpose is to select the features which more easily
    divide the target space by fitting a simple tree (usually 3-4 level depth)
    for each feature.

    Args:
        task (string): 'classification' or 'regression'
        max_depth (int): Maximum depth of the fitted trees. Default is three.
        metric (function): Sklearn style metric which accepts (y_true, y_pred) parameters.
            As default, mse will be used for regression task and accuracy for classification.
        use_proba (bool): Whether the passed metric expects probabilities or labels.
            Default is False (Trees will use labels). This is only used for classifiction tasks.
        lower_better(bool): The columns with a lower asscociated metric will be considered more
            important (usually when chosing metrics like mse, mae...). For classification tasks
            this should be False. Default = True.
        tree_kwargs (dict): .

    """

    def __init__(self, task, max_depth=3,
                 metric=None,
                 use_proba=False,
                 lower_better=True,
                 tree_kwargs={}):
        """Constructor"""
        if task != 'classification' and task != 'regression':
            raise ValueError("Choose one of [classification, regression] task")

        if metric is None:
            metric = mean_squared_error if task == 'regression' else accuracy_score

        if metric == accuracy_score and lower_better:
            lower_better = False

        self.task = task
        self.max_depth = max_depth
        self.metric = metric
        self.use_proba = use_proba
        self.tree_kwargs = tree_kwargs
        self.lower_better = lower_better

    def _tree_importance(self, feature, target, random_state=42):
        """ Fits a tree to the feature, trying to predict the target.

        Args:
            feature (np.array): with the feature values.
            target (np.array): target to predict

        Returns:
            The obtained metric for the fit.
        """
        if self.task == 'classification':
            tree = DecisionTreeClassifier(max_depth=self.max_depth,
                                          random_state=random_state,
                                          **self.tree_kwargs).fit(feature, target)

            if self.use_proba:
                p = tree.predict_proba(feature)
            else:
                p = tree.predict(feature)

            # Needed to suppress warnings in case a tree only predicts one class.
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                res = self.metric(target, p)
        else:
            tree = DecisionTreeRegressor(max_depth=self.max_depth,
                                         random_state=random_state,
                                         **self.tree_kwargs).fit(feature, target)
            p = tree.predict(feature)
            res = self.metric(target, p)
        return res

    def fit(self, X, y, tree_random_state=42) -> "TreeSelector":
        """ Fits the estimator and calculates the features importances

        Args:
            X (np.array or pd.DataFrame): Features
            y (np.array or pd.Series): Target variable
            reverse (bool): Whether to reverse the order of the result

        Returns:
            (dict): dictionary with two keys:

                - features: Ordered feature indices by the obtained metric\n
                - metrics: value of the metric for each feature.
            (TreeSelector): the estimator itself
        """
        num_features = X.shape[-1]
        errors = []
        for feat in range(num_features):
            err = self._tree_importance(np.array(X)[:, feat].reshape(-1, 1), y,
                                        random_state=tree_random_state)
            errors.append(err)

        # Sorted feature indexes
        self.ordered_feat_idx_ = np.argsort(errors) if self.lower_better else np.argsort(errors)[::-1]

        # Sorted feature errors (or the chosen metric)
        self.ordered_feat_importances_ = [errors[e] for e in self.ordered_feat_idx_]

        return self

    def transform(self, X, keep_top_n=None):
        check_is_fitted(self, ["ordered_feat_idx_", "ordered_feat_importances_"])
        filtered_data = None
        if type(X) == "pd.DataFrame":
            filtered_data = X.iloc[:, self.ordered_feat_idx_[:keep_top_n]]
        else:
            filtered_data = X[:, self.ordered_feat_idx_[:keep_top_n]]

        return filtered_data

    @property
    def feature_importances(self):
        return dict(features=self.ordered_feat_idx_, metrics=self.ordered_feat_importances_)


from .basetests import RobustDataTest
from .errors import FailedTestError
from typing import Union, Any, List, Callable, Dict, Tuple
from sklearn.preprocessing import OneHotEncoder, LabelEncoder
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.compose import ColumnTransformer
import psutil

from mercury.dataschema import DataSchema
from mercury.dataschema.feature import DataType, FeatType

import numpy as np
import pandas as pd


class SameSchemaTest(RobustDataTest):
    """
    This test ensures a certain dataset, or new batch of samples, follows a
    previously defined schema.

    On the `run` call, it will make a schema validation. If something isn't right, it
    will raise an Exception, crashing the test.

    Args:
        base_dataset: Dataset which will be evaluated
        schema_ref: Schema the base_dataset will be evaluated against. It can be either a
                    `DataSchema` object or a string. In case of the later, it must be a
                    path to a previously serialized schema.
        custom_feature_map: Dictionary with <column_name: FeatType>. You can provide this in case
                            the DataSchema wrongly infers the types for the new data batch. If
                            this is None, the types will be auto inferred.
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.dataschema import DataSchema
        >>> schma_reference = DataSchema().generate(df_train).calculate_statistics()
        >>> from mercury.robust.data_tests import SameSchemaTest
        >>> test = SameSchemaTest(df_inference, schma_reference)
        >>> test.run()
        ```

    """
    def __init__(self,
                 base_dataset: "pandas.DataFrame",  # noqa: F821
                 schema_ref: Union[str, DataSchema],
                 custom_feature_map: Dict[str, FeatType] = None,
                 name: str = None,
                 *args: Any, **kwargs: Any):
        super().__init__(base_dataset, name, *args, **kwargs)
        self.schema_ref = schema_ref
        self.custom_feature_map = custom_feature_map

        if type(schema_ref) == str:
            self.schema_ref = DataSchema.load(schema_ref)

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if any of the checks fail.
        """
        super().run(*args, **kwargs)
        current_schema = DataSchema().generate(self.base_dataset, force_types=self.custom_feature_map)
        try:
            self.schema_ref.validate(current_schema)
        except RuntimeError as e:
            raise FailedTestError(str(e))


class LinearCombinationsTest(RobustDataTest):
    """
    This test ensures a certain dataset doesn't have any linear combination between
    its numerical columns and no categorical variable is redundant. See the following functions if you want to know further details:

      - For numerical features: _lin_combs_in_columns.lin_combs_in_columns
      - For categorical features: _CategoryStruct.individually_redundant

    In case any combination or redundancy is detected, the test fails.

    Args:
        base_dataset: Dataset which will be evaluated
        schema_custom_feature_map: Internally, this test generates a DataSchema object. In case you find it makes
                                   wrong feature type assignations to your features you can pass here a dictionary which
                                   specify the feature type of the columns you want to fix. (See DataSchema. `force_types`
                                   parameter for more info on this).
        dataset_schema: Pre built schema. This argument is complementary to `schema_custom_feature_map`. In case you want to
                        manually build your own DataSchema object, you can pass it here and the test will internally use it
                        instead of the default, automatically built. If you provide this parameter, `schema_custom_feature_map`
                        will not be used.
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.data_tests import LinearCombinationsTest
        >>> test = LinearCombinationsTest(df_train)
        >>> test.run()
        ```
    """
    def __init__(self,
                 base_dataset: "pandas.DataFrame",  # noqa: F821
                 schema_custom_feature_map: Dict[str, FeatType] = None,
                 dataset_schema: DataSchema = None,
                 name: str = None, *args: Any, **kwargs: Any):
        super().__init__(base_dataset, name, *args, **kwargs)
        self._schema_custom_feature_map = schema_custom_feature_map
        self._dataset_schema = dataset_schema

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if any of the checks fail.
        """
        from ._linalg import lin_combs_in_columns
        from ._category_struct import CategoryStruct

        super().run(*args, **kwargs)

        if self._dataset_schema is None:
            current_schema = DataSchema().generate(self.base_dataset, force_types=self._schema_custom_feature_map)
        else:
            current_schema = self._dataset_schema

        lin_combinations = None
        numeric_feats = current_schema.continuous_feats + current_schema.discrete_feats
        if len(numeric_feats) > 0:
            cont_cols = self.base_dataset.loc[:, numeric_feats].values

            lin_combinations = lin_combs_in_columns(
                np.matmul(cont_cols.T, cont_cols)  # Compress original matrix
            )

        if lin_combinations is not None:
            # Create message with the linear combinations found
            lin_combinations = np.around(lin_combinations, decimals=5)
            lin_combs_found = []
            for i in range(lin_combinations.shape[0]):
                lin_comb_idx = np.where(lin_combinations[i] != 0)[0]
                lin_comb_cols = self.base_dataset.loc[:, numeric_feats].columns[lin_comb_idx].tolist()
                lin_combs_found.append(lin_comb_cols)
            raise FailedTestError(f"Test failed. Linear combinations for continuous features were encountered: {lin_combs_found}.")

        individually_redundant = CategoryStruct.individually_redundant(self.base_dataset, current_schema.categorical_feats)
        if len(individually_redundant) > 0:
            raise FailedTestError((
                f"""Test failed. Any of these categorical variables is redundant: {individually_redundant}. """
                """Try deleting any of them."""
            ))


class DriftTest(RobustDataTest):
    """
    This test ensures the distributions of new dataset, or new batch of data, are
    not too different from a reference dataset (i.e. training data). In other words, it
    checks for no data drift between features.

    If drift is detected on any of the features, it will raise an Exception, crashing the test.

    Args:
        base_dataset: Dataset which will be evaluated
        schema_ref: Schema the base_dataset will be evaluated against. It can be either a
                    `DataSchema` object or a string. In case of the later, it must be a
                    path to a  previously serialized schema.
        drift_detector_args: Dictionary with arguments passed to the internal drift detectors:
                            HistogramDistanceDrift for continuous features and Chi2Drift for
                            the categoricals.
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.dataschema import DataSchema
        >>> schma_reference = DataSchema().generate(df_train).calculate_statistics()
        >>> from mercury.robust.data_tests import DriftTest
        >>> test = DriftTest(df_inference, schma_reference)
        >>> test.run()
        >>> # Check test result details
        >>> test.info()
        ```
    """
    def __init__(self,
                 base_dataset: "pandas.DataFrame",  # noqa: F821
                 schema_ref: Union[str, DataSchema] = None,
                 drift_detector_args: dict = None,
                 name: str = None,
                 *args: Any, **kwargs: Any):
        super().__init__(base_dataset, name, *args, **kwargs)
        self.schema_ref = schema_ref

        # Default args for ALL detectors. Depending on the implementation, only the
        # necessary ones will be used.
        self._detector_args = dict(
            distance_metric="hellinger",
            correction="bonferroni",
            n_runs=3,
            test_size=0.3,
            p_val=0.01,
            n_permutations=100
        )

        if isinstance(drift_detector_args, dict):
            self._detector_args.update(drift_detector_args)

        if type(schema_ref) == str:
            self.schema_ref = DataSchema.load(schema_ref)

        self.continuous_drift_metrics = None
        self.cat_drift_metrics = None

    def info(self):
        if (self.continuous_drift_metrics is None) and (self.cat_drift_metrics is None):
            return None

        dict_info = {}
        if self.continuous_drift_metrics is not None:
            dict_info["continuous_drift_metrics"] = self.continuous_drift_metrics["score"]
        if self.cat_drift_metrics is not None:
            dict_info["cat_drift_metrics"] = self.cat_drift_metrics["score"]

        return dict_info

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if any drift is detected
        """
        from mercury.monitoring.drift.histogram_distance_drift_detector import HistogramDistanceDrift
        from mercury.monitoring.drift.chi2_drift_detector import Chi2Drift

        super().run(*args, **kwargs)

        schema_ref = self.schema_ref

        # Test numerical features with HistogramDistance
        featnames = schema_ref.continuous_feats + schema_ref.discrete_feats

        if len(featnames) > 0:

            src_histograms, tgt_histograms = self._get_features_histograms(featnames)

            detector_args = {
                'distance_metric': self._detector_args['distance_metric'],
                'correction': self._detector_args['correction'],
                'p_val': self._detector_args['p_val'],
                'n_permutations': self._detector_args['n_permutations']
            }

            drift_detector = HistogramDistanceDrift(
                distr_src=src_histograms,
                distr_target=tgt_histograms,
                features=featnames,
                **detector_args
            )

            self.continuous_drift_metrics = drift_detector.calculate_drift()

            if self.continuous_drift_metrics['drift_detected']:
                raise FailedTestError(f"Test failed. Drift was detected on the following features: {drift_detector.get_drifted_features()}")

        # Test categorical feats with Chi2
        featnames = schema_ref.binary_feats + schema_ref.categorical_feats
        if len(featnames) > 0:

            src_histograms, tgt_histograms = self._get_features_histograms(featnames)

            chi2_args = {
                'correction': self._detector_args['correction'],
                'p_val': self._detector_args['p_val'],
            }

            drift_detector = Chi2Drift(
                distr_src=src_histograms,
                distr_target=tgt_histograms,
                features=featnames,
                **chi2_args
            )

            self.cat_drift_metrics = drift_detector.calculate_drift()

            if self.cat_drift_metrics['drift_detected']:
                raise FailedTestError(f"Test failed. Drift was detected on the following features: {drift_detector.get_drifted_features()}")

    def _get_features_histograms(self, featnames: list) -> Tuple[List[np.array], List[np.array]]:
        """
        Returns the source and target histograms of each feature in featnames

        Args:
            featnames: list of feature names to obtain the histograms
        Returns
            Two list src_histograms and tgt_histograms, with the first containing the histograms of the source dataset and
            the latter containing the histograms of the target dataset

        """

        src_histograms = []
        tgt_histograms = []
        for feat in featnames:
            src_histogram, tgt_histogram = self._get_histograms(feat)
            src_histograms.append(src_histogram)
            tgt_histograms.append(tgt_histogram)
        return src_histograms, tgt_histograms

    def _get_histograms(self, feat: str) -> Tuple[np.array, np.array]:
        """
        Returns histograms for a specific feature, feat.
        Obtains the histogram for the reference dataset using the dataschema,
        and the histogram for the target dataset from the base_dataset.

        Args:
            feat: feature name

        Returns:
            Two numpy arrays, the first containing the histogram for the source dataset,
            and the second containing the histogram for the target dataset

        Raises:
            ValueError if the new feature has a new value not present in the schema.
        """

        schema_ref = self.schema_ref
        # Obtain source histogram from dataschema stats
        src_hist = np.array(schema_ref.feats[feat].stats['distribution'])

        # Obtain target histogram, depending on if its a discrete feature or a numeric one
        discrete_feats = self.schema_ref.categorical_feats + self.schema_ref.binary_feats

        if feat in discrete_feats:
            # If features are discrete, we don't need to take numerical ranges into
            # account
            domain = self.base_dataset.loc[:, feat].value_counts()
            values = set(domain.index)
            reference_values = set(schema_ref.feats[feat].stats['distribution_bins'])

            # If new dataset has categories that base doesn't have, raise FailedTestError
            new_values = values - reference_values
            if len(new_values) > 0:
                raise FailedTestError(
                    f"Test failed. Drift was detected on feature '{feat}' because has new categories: {new_values}"
                )

            # If new dataset has missing categories than reference, we add them with 0 count
            missing_values = reference_values - values
            if len(missing_values) > 0:
                for v in missing_values:
                    domain[v] = 0

            domain.index = domain.index.astype(str)
            str_bins = [str(x) for x in schema_ref.feats[feat].stats['distribution_bins']]
            domain = domain[str_bins]
            tgt_hist = domain.values

        else:
            # Feature is  numerical, we need to match source and target bins
            tgt_hist = np.histogram(self.base_dataset.loc[:, feat],
                                    bins=schema_ref.feats[feat].stats['distribution_bins']
                                    )[0]

            # If we have a new minimum, then we will have target samples outside the limits of the source distribution
            # We add a new bin at the beginning of both histograms for these samples
            if self.base_dataset.loc[:, feat].min() < schema_ref.feats[feat].stats['min']:
                source_samples_new_bin = 0
                target_samples_new_bin = (self.base_dataset.loc[:, feat] < schema_ref.feats[feat].stats['min']).sum()
                src_hist = np.insert(src_hist, 0, source_samples_new_bin)
                tgt_hist = np.insert(tgt_hist, 0, target_samples_new_bin)

            # If we have a new maximum, then we will have target samples outside the limits of the source distribution
            # We add a new bin at the end of both histograms for these samples
            if self.base_dataset.loc[:, feat].max() > schema_ref.feats[feat].stats['max']:
                source_samples_new_bin = 0
                target_samples_new_bin = (self.base_dataset.loc[:, feat] > schema_ref.feats[feat].stats['max']).sum()
                src_hist = np.insert(src_hist, len(src_hist), source_samples_new_bin)
                tgt_hist = np.insert(tgt_hist, len(tgt_hist), target_samples_new_bin)

            # Convert counts to densities
            tgt_hist = tgt_hist / tgt_hist.sum()

        # Trick because Drift detectors don't work with densities yet
        src_hist *= 1000
        tgt_hist *= 1000

        return src_hist, tgt_hist


class LabelLeakingTest(RobustDataTest):
    """
    This test ensures the target variable is not being leaked into the predictors
    (i.e. no feature has a strong relationship with the target). For this, it
    uses the TreeSelector class which tries to predict the
    target variable given each one of the features. If, given a particular feature, the
    target is easily predicted, it means that feature is very important. By default,
    performance is measured by ROC-AUC for classification problems and R^2 for regression.

    After the test has been executed, two attributes are made available for further inspection.

        1) `importances_`: dictionary with the the feature importances. That is,
           the closer a feature is to 1 (ideal value), the most important it is. If any
           feature f is f > 1 - threshold, the test will fail.
        2) `_selector`: Fitted TreeSelector used for the test.

    Args:
        base_dataset: Dataset which will be evaluated. Features must be numerical (i.e. no strings/objects)
        label_name: column which represent the target label. If it's multi-categorical, it
                    must be label-encoded.
        task: Task of the dataset. It must be either 'classification' or 'regression'. If None provided
              it will be auto inferred from the `label_name` column.
        threshold: Custom threshold which will be used when considering a particular feature important.
                   If the obtained score for a feature < threshold, it will be considered "too important", and, thus,
                   the test will fail.
        metric: Metric to be used by the internal TreeSelector. If None, ROC AUC will be used for classification and
                R2 for regression.
        ignore_feats: Features which will be not tested.
        schema_custom_feature_map: Internally, this test generates a DataSchema object. In case you find it makes
                                   wrong feature type assignations to your features you can pass here a dictionary which
                                   specify the feature type of the columns you want to fix. (See DataSchema. `force_types`
                                   parameter for more info on this).
        dataset_schema: Pre built schema. This argument is complementary to `schema_custom_feature_map`. In case you want to
                        manually build your own DataSchema object, you can pass it here and the test will internally use it
                        instead of the default, automatically built. If you provide this parameter, `schema_custom_feature_map`
                        will not be used.
        handle_str_cols (str): if the dataset contains string columns, this parameter indicates how to handle them. If 'error'
            then a ValueError exception is raised if the dataset contains string columns. If 'transform' the string columns
            are transformed to integer using a LabelEncoder. Default value is 'error'
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.data_tests import LabelLeakingTest
        >>> test = LabelLeakingTest(
        >>>     train_df,
        >>>     label_name = "y",
        >>>     task = "classification",
        >>>     threshold = 0.05,
        >>> )
        >>> test.run()
        >>> # Check test result details
        >>> test.info()
        ```
    """
    def __init__(self,
                 base_dataset: "pandas.DataFrame",  # noqa: F821
                 label_name: str,
                 task: str = None,
                 threshold: float = None,
                 metric: Callable = None,
                 ignore_feats: List[str] = None,
                 schema_custom_feature_map: Dict[str, FeatType] = None,
                 dataset_schema: DataSchema = None,
                 handle_str_cols: str = "error",
                 name: str = None,
                 *args: Any, **kwargs: Any):
        super().__init__(base_dataset, name, *args, **kwargs)

        self.label_name = label_name
        self.task = task
        self.threshold = threshold
        self.ignore_feats = ignore_feats
        self.metric = metric
        self.importances_ = None
        self._schema_custom_feature_map = schema_custom_feature_map
        self._base_schema = dataset_schema
        self.handle_str_cols = handle_str_cols

    def info(self):
        if self.importances_ is not None:
            return {"importances": self.importances_}
        else:
            return None

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if any of the checks fail.
        """
        from ._treeselector import TreeSelector
        from sklearn.metrics import r2_score, roc_auc_score

        super().run(*args, **kwargs)

        base_dataset = self.base_dataset

        if self.ignore_feats is not None:
            base_dataset = base_dataset.loc[:, [c for c in base_dataset.columns if c not in self.ignore_feats]]

        label_name = self.label_name
        threshold = self.threshold
        metric = self.metric

        if self._base_schema is None:
            schema = DataSchema().generate(base_dataset, force_types=self._schema_custom_feature_map)
        else:
            schema = self._base_schema

        if not self.task:
            if label_name in schema.binary_feats or label_name in schema.categorical_feats:
                self.task = 'classification'
            if label_name in schema.discrete_feats or label_name in schema.continuous_feats:
                self.task = 'regression'

        task = self.task

        if task not in ('classification', 'regression'):
            raise ValueError("Error. 'task' must be either classification or regression")

        # This threshold has been found to be working well in most tested cases for
        # R2 and AUC.
        threshold = 0.05 if not threshold else threshold
        if task == 'classification':
            metric = roc_auc_score if metric is None else metric
        else:
            metric = r2_score if metric is None else metric

        # The TreeSelector used later doesn't support string columns. Raise error or transform them
        str_cols = schema.get_features_by_type(datatype=DataType.STRING)
        if len(str_cols) > 0:
            if self.handle_str_cols.lower() == 'error':
                raise ValueError(
                    "String column found. Set 'handle_str_cols' parameter to 'transform' to automatically transform "
                    "string columns to integers or transform them before using this test."
                )
            elif self.handle_str_cols.lower() == 'transform':
                # We need to create a copy in this case to avoid to modify the original dataset
                base_dataset = base_dataset.copy()
                for col in str_cols:
                    if col in base_dataset.columns:
                        base_dataset.loc[:, col] = LabelEncoder().fit_transform(base_dataset.loc[:, col])
            else:
                raise ValueError("Wrong value for 'handle_str_cols' parameter. Set to 'error' or 'transform'")

        # Separate features from target
        X = base_dataset.loc[:, [f for f in base_dataset.columns if f != label_name]]
        y = base_dataset.loc[:, label_name]

        selector = TreeSelector(task, metric=metric)
        selector.fit(X, y)

        # Store selector for inspection. Mainly debugging purposes
        self._selector = selector

        # We calculate our "importance" scores depending on whether we have a
        # classification or regression dataset.
        importances = selector.feature_importances

        # By default,  we ensure the best performing feature (i.e.
        # the one with highest metric) is not too close to one (the best
        # possible value for classification and regression / AUC or R^2).
        metrics = 1 - np.clip(np.array(importances['metrics']), 0, 1)
        high_importance_feats = metrics < threshold

        # Store computed importances for post-analysis
        ids = X.columns[selector.feature_importances['features']].tolist()
        self.importances_ = dict()
        for i, val in enumerate(ids):
            self.importances_[val] = 1 - metrics[i]  # Recover original metric instead of its inverse

        # If any of the features is too important the test fails.
        if high_importance_feats.any():
            idxs = selector.feature_importances['features'][high_importance_feats]
            names = X.columns[idxs].tolist()
            raise FailedTestError((
                f"Test failed because high importance features were detected: {names}. "
                "Check for possible target leaking."
            ))


class NoisyLabelsTest(RobustDataTest):
    """
    This test looks if the labels of a dataset contain a high level of noise. Internally, it uses the cleanlab
    library to obtain those samples in the dataset where the labels are considered as noisy. Noisy labels can happen because
    the sample is incorrectly labelled, because the sample could belong to several label categories, or some other reason.
    If the percentage of samples with noisy labels is higher than a specified threshold, then the test fails.
    It can be used with tabular datasets and with text datasets. In the case of a text dataset, the argument `text_col` must
    be specified.
    This test only works for classification tasks.

    After the test has been executed, two attributes are made available for further inspection.

        1) `idx_issues_`: indices of the labels detected as issues (only available when `calculate_idx_issues` is True)
        2) `rate_issues_`: percentage of labels detected containing possible issues.

    IMPORTANT: If the Test reports convergence problems or it takes too long, then you can change the model used in the
    algorithm (see example below). Sometimes, the default Logistic Regression used might not converge in some datasets.
    In those cases, changing the solver is usually enough to solve the problem.

    Args:
        base_dataset (pandas.DataFrame): Dataset which will be evaluated.
        label_name (str): column which represents the target label.
        threshold (float): threshold to specify the percentage of noisy labels from which the test will fail. Default value is 0.4
        text_col (str): column containing the text. Only has to be specified when using a text dataset.
        preprocessor (Union[TransformerMixin, BaseEstimator]): Optional preprocessor for the features. If not specified, it will
            create a preprocessor with OneHotEncoder to encode categorical features in case of tabular data
            (text_col not specified) or a CountVectorizer to encode text in case of a text dataset(text_col specified)
        ignore_feats (List[str]): features that will not be used in the algorithm to detect noisy labels.
        calculate_idx_issues (bool): whether to calculate also the index of samples with label issues. If True, the idx will be
            available in `idx_issues_` attribute. Default value is True.
        label_issues_args (dict): arguments for the algorithm to detect noisy labels. You can check documentation from
            `_label_cleaning.get_label_issues` to see all available arguments.
        schema_custom_feature_map: Internally, this test generates a DataSchema object. In case you find it makes
                                   wrong feature type assignations to your features you can pass here a dictionary which
                                   specify the feature type of the columns you want to fix. (See DataSchema. `force_types`
                                   parameter for more info on this).
        dataset_schema: Pre built schema. This argument is complementary to `schema_custom_feature_map`. In case you want to
                        manually build your own DataSchema object, you can pass it here and the test will internally use it
                        instead of the default, automatically built. If you provide this parameter, `schema_custom_feature_map`
                        will not be used.
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.data_tests import NoisyLabelsTest
        >>> test = NoisyLabelsTest(
        >>>     base_dataset=train_df,
        >>>     label_name="label_col",
        >>>     text_col = "text_col,
        >>>     threshold = 0.2,
        >>>     label_issues_args={"clf": LogisticRegression(solver='sag')}
        >>> )
        >>> test.run()
        >>> # Percentage of samples with possible label issues
        >>> test.rate_issues_
        >>> # Access to samples with possible label issues
        >>> train_df.iloc[test.idx_issues_]
        ```
    """

    def __init__(self,
                 base_dataset: "pandas.DataFrame",  # noqa: F821
                 label_name: str,
                 threshold: float = None,
                 text_col: str = None,
                 preprocessor: Union["TransformerMixin", "BaseEstimator"] = None,  # noqa: F821
                 ignore_feats: List[str] = None,
                 calculate_idx_issues: bool = True,
                 label_issues_args: dict = None,
                 name: str = None,
                 schema_custom_feature_map: Dict[str, FeatType] = None,
                 dataset_schema: DataSchema = None,
                 *args: Any, **kwargs: Any):
        super().__init__(base_dataset, name, *args, **kwargs)

        self.label_name = label_name
        self.threshold = 0.4 if not threshold else threshold
        self.preprocessor = preprocessor
        self.text_col = text_col
        self.ignore_feats = ignore_feats
        self.calculate_idx_issues = calculate_idx_issues
        self.rate_issues_ = None
        self._schema_custom_feature_map = schema_custom_feature_map
        self._dataset_schema = dataset_schema

        # Default args for function to get label issues
        self.label_issues_args = dict(
            clf=None,
            n_folds=5,
            frac_noise=1.0,
            num_to_remove_per_class=None,
            prune_method='prune_by_noise_rate',
            sorted_index_method=None,
            n_jobs=None,
            seed=None
        )

        if isinstance(label_issues_args, dict):
            self.label_issues_args.update(label_issues_args)

    def _preprocess_features_from_text(self, X):
        """
        Converts text from text_col column in X to a feature vector.

        Args:
            X: features dataframe containing the text to preprocess

        Returns:
            np.array with the preprocessed text
        """

        # If preprocessor specified we use it, otherwise create a simple CountVectorizer
        if self.preprocessor:
            preprocessor = self.preprocessor
        else:
            preprocessor = CountVectorizer(max_features=10000).fit(X[self.text_col])
        X = preprocessor.fit_transform(X[self.text_col])

        return X

    def _preprocess_features_from_tabular(self, X, current_schema):
        """
        Preprocesses the features from a tabular dataframe.

        Args:
            X: features dataframe to preprocess
            current_schema: schema of the dataset

        Returns:
            np.array with the preprocessed features
        """
        # If preprocessor specified we use it
        if self.preprocessor:
            preprocessor = self.preprocessor
            X = preprocessor.fit_transform(X)
        else:
            # If preprocessor not specified, create one to preprocess string features with OneHotEncoder
            str_feats = [c for c in current_schema.get_features_by_type(DataType.STRING) if c in X.columns]
            cat_feats = list(set.union(set(str_feats), set(current_schema.categorical_feats)))
            if len(cat_feats) > 0:
                preprocessor = ColumnTransformer(
                    transformers=[('ohe', OneHotEncoder(handle_unknown='ignore'), cat_feats)],
                    remainder='passthrough'
                )
                X = preprocessor.fit_transform(X)
            else:
                X = X.values
        return X

    def _preprocess_features(
        self,
        X: "pandas.DataFrame",  # noqa: F821
        current_schema: DataSchema
    ) -> np.array:
        """
        Preprocesses the features dataframe to create a np.array ready for the label issues detection
        algorithm.

        Args:
            X: features dataframe to preprocess
            current_schema: schema of the dataset

        Returns:
            np.array with the preprocessed features
        """

        # If text_col is specified, we have to preprocess (only) the text column
        if self.text_col:
            X = self._preprocess_features_from_text(X)
        else:
            X = self._preprocess_features_from_tabular(X, current_schema)

        return X

    def info(self):
        if self.rate_issues_ is not None:
            return {"rate_issues": self.rate_issues_}
        else:
            return None

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if noise above threshold is detected
        """
        from ._label_cleaning import get_label_issues, get_confident_joint

        super().run(*args, **kwargs)

        base_dataset = self.base_dataset

        if self.ignore_feats is not None:
            base_dataset = base_dataset.loc[:, [c for c in base_dataset.columns if c not in self.ignore_feats]]

        if self._dataset_schema is None:
            current_schema = DataSchema().generate(base_dataset, force_types=self._schema_custom_feature_map)
        else:
            current_schema = self._dataset_schema

        # Separate features from target
        X = base_dataset.loc[:, [f for f in base_dataset.columns if f != self.label_name]]
        y = base_dataset.loc[:, self.label_name].values

        # Preprocess features
        X = self._preprocess_features(X, current_schema)

        if self.calculate_idx_issues:
            # Get label issues
            self.idx_issues_, _ = get_label_issues(
                X=X,
                y=y,
                **self.label_issues_args
            )

            # Obtain number of label issues
            if self.idx_issues_.dtype == bool:
                self.num_issues = self.idx_issues_.sum()
            else:
                self.num_issues = len(self.idx_issues_)

        else:
            # Compute only confident joint and calculate number of issues as the sum of non-diagonal elements
            confident_join = get_confident_joint(
                X, y, n_folds=self.label_issues_args["n_folds"], clf=self.label_issues_args["clf"], seed=self.label_issues_args["seed"]
            )
            self.num_issues = confident_join.sum() - np.trace(confident_join)

        # Calculate percentage of obtained labels with issues
        self.rate_issues_ = self.num_issues / len(base_dataset)

        # If percentage of issues is higher than threshold, throw exception
        if self.rate_issues_ > self.threshold:
            raise FailedTestError(
                f"Test failed. High level of noise detected in labels. Percentage of labels with noise: {self.rate_issues_}"
            )


class CohortPerformanceTest(RobustDataTest):
    """
    This test looks if some metric performs poorly for some cohort of your data when compared with other groups.
    The computed metric is specified with the argument `eval_group_fn`, which the user needs to define (see example).
    The metric is computed for all the groups of the variable specified with the variable `group_col`.
    When the argument `compare_max_diff` is True, then the comparison is between the group which has the maximum
    metric and the group which has the minimum metric. If the difference is higher than the specified threshold
    then the test will fail. When the argument `compare_max_diff` is False, then the metric calculated for each
    group is compared against the mean of the whole dataset. If the difference of any of the groups with the mean
    is higher than the threshold, then the test will fail.
    The argument `threshold_is_percentage` controls if the threshold is compared with the absolute value of the
    difference (`threshold_is_percentage=False`), or if is compared with the percentage of the difference
    (`threshold_is_percentage=True`).

    Args:
        base_dataset (pandas.DataFrame): Dataset which will be evaluated. It must contain the specified `group_col`
            and the necessary columns to calculate the metric defined in `eval_group_fn`
        group_col (str): column name which contains the groups to be evaluated
        eval_fn (Callable): evaluation function which computes the metric to evaluate. It must return a float
        threshold (float): threshold to compare. If `compare_max_diff` is True and the difference between the
            maximum metric and minimum metric is higher than the threshold, then the test fails. If
            `compare_max_diff` is False and the difference between the mean metric of the dataset and the
            metric in any of the groups is higher than the threshold, then the test fails.
        compare_max_diff (bool): If True, then the comparison is between the group which has the max metric and the
            group which has the min metric. If False, then the comparison is between the mean of the whole dataset
            and all the other groups. Default value is True
        threshold_is_percentage (bool): If True, then the comparison with the threshold is with the percentage
            of the computed differences, therefore the threshold represents the percentage.
            If False, then the comparison with the threshold is with the absolute values of the differences,
            therefore the threshold represents the absolute value. Default value is True.
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.data_tests import CohortPerformanceTest
        >>> def eval_acc(df):
        >>>     return accuracy_score(df["y_true"], df["y_pred"])
        >>> test1 = CohortPerformanceTest(
        >>>     base_dataset=dataset, group_col="gender", eval_fn = eval_acc, threshold = 0.2, compare_max_diff=True
        >>> )
        >>> test1.run()
        >>> # Check test result details
        >>> test1.info()
        ```
    """

    def __init__(
        self,
        base_dataset: "pandas.DataFrame",  # noqa: F821
        group_col: str,
        eval_fn: Callable,
        threshold: float,
        compare_max_diff: bool = True,
        threshold_is_percentage: bool = True,
        name: str = None,
        *args: Any,
        **kwargs: Any
    ):
        super().__init__(base_dataset, name, *args, **kwargs)
        self.group_col = group_col
        self.eval_fn = eval_fn
        self.threshold = threshold
        self.compare_max_diff = compare_max_diff
        self.threshold_is_percentage = threshold_is_percentage
        self.metric_by_group = None

    def info(self):
        if self.metric_by_group is not None:
            return {"metric_by_group": self.metric_by_group}
        else:
            return None

    def _compare_max_diff(self):
        max_value = self.metric_by_group.max()
        min_value = self.metric_by_group.min()
        if self.threshold_is_percentage:
            self.diff = (max_value - min_value) / min_value
        else:
            self.diff = max_value - min_value
        if self.diff > self.threshold:
            raise FailedTestError((
                f"Test failed because max difference {round(self.diff,3)} is higher than threshold {self.threshold}"
                f" for groups in {self.group_col}. "
                f" Group {self.metric_by_group.idxmax()} has a metric of {round(max_value,3)}"
                f" and Group {self.metric_by_group.idxmin()} has a metric of {round(min_value,3)}"
            ))

    def _compare_diff_with_mean(self):
        self.mean_metric = self.eval_fn(self.base_dataset)
        if self.threshold_is_percentage:
            self.diff = (self.metric_by_group - self.mean_metric) / self.mean_metric
        else:
            self.diff = (self.mean_metric - self.metric_by_group)
        mean_diff_higher = self.diff.abs() > self.threshold
        if any(mean_diff_higher):
            raise FailedTestError((
                f"Test failed because the metric difference in "
                f"groups {self.metric_by_group[mean_diff_higher].index.values} "
                f"is higher than threshold with respect the mean in the dataset"
            ))

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if metric difference above the threshold is detected
        """
        super().run(*args, **kwargs)

        self.metric_by_group = self.base_dataset.groupby(self.group_col).apply(lambda x: self.eval_fn(x))

        if self.compare_max_diff:
            self._compare_max_diff()
        else:
            self._compare_diff_with_mean()


class SampleLeakingTest(RobustDataTest):
    """
    This test looks if there are samples in the test dataset that are identical to samples in the base/train dataset.
    It considers that a sample is the same if it contains the same values for all the columns. If the number of duplicated
    samples is higher than the allowed by the `threshold` then the test fails.

    Args:
        base_dataset (pandas.DataFrame): Training/Base dataset.
        test_dataset (pandas.DataFrame): Test dataset that we look if we have the same samples than in the
            `base_dataset`. It must contain the same columns as the `base_dataset`
        ignore_feats (List): List of the name of the columns that we want to ignore when comparing the samples
        threshold (Union[float, int]): max percentage (float) or number (int) of samples that we allowed to be duplicated.
            By default is 0
        use_hash (bool): If True, it will create a hash for the rows in the dataframes in order to find duplicates. If False
            it will use `duplicate()` pandas method. Using hash usually results in faster execution in bigger datasets. If
            None it will use one method or other depending of the available memory
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.data_tests import SampleLeakingTest
        >>> test = SampleLeakingTest(base_dataset=my_train_dataframe, test_dataset=my_test_dataset)
        >>> test.run()
        >>> # Check test result details
        >>> test.info()
        ```
    """

    def __init__(
        self,
        base_dataset: "pandas.DataFrame",  # noqa: F821
        test_dataset: "pandas.DataFrame",  # noqa: F821
        ignore_feats: List[str] = None,
        threshold: Union[int, float] = 0,
        name: str = None,
        use_hash: bool = None,
        *args: Any, **kwargs: Any
    ):
        super().__init__(base_dataset, name, *args, **kwargs)
        self.test_dataset = test_dataset
        self.ignore_feats = ignore_feats if ignore_feats else []
        self.threshold = threshold
        self.use_hash = use_hash
        self._is_duplicated_test = None

    def info(self):

        if self._is_duplicated_test is None:
            return None

        return {
            'num_duplicated': self._is_duplicated_test.sum(),
            'percentage_duplicated': self._is_duplicated_test.sum() / len(self.test_dataset)
        }

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if samples in test set existing in train are above threshold
        """
        self._verify_datasets()
        self._is_duplicated_test = self._find_sample_leaking()

        if self._high_number_of_duplicates():
            raise FailedTestError(
                f"Num of samples in test set that appear in train set is {self._is_duplicated_test.sum()} "
                f"(a proportion of {round(self._is_duplicated_test.sum() / len(self.test_dataset), 3)} test samples )"
                f"and the max allowed is {self.threshold}")

    def _find_sample_leaking(self):

        if self.use_hash is None:
            self.use_hash = self._infer_use_hashing()

        if self.use_hash:
            return self._find_sample_leaking_hash_impl()
        else:
            return self._find_sample_leaking_duplicated_impl()

    def _infer_use_hashing(self):
        if psutil.virtual_memory().percent > 20.0:
            return True
        else:
            return False

    def _find_sample_leaking_duplicated_impl(self):

        # Concatenate the datasets without duplicates inside of the same sets
        cols = [c for c in self.base_dataset.columns if c not in self.ignore_feats]
        base_dataset_no_duplicates = self.base_dataset[cols].drop_duplicates()
        test_dataset_no_duplicates = self.test_dataset[cols].drop_duplicates()
        concat_dataset = pd.concat([base_dataset_no_duplicates, test_dataset_no_duplicates])

        # Find the duplicates in the concatenated dataset
        is_duplicated_concat = concat_dataset.duplicated(keep=False).values

        # Return boolean array of elements in test set that appear in train
        return is_duplicated_concat[len(base_dataset_no_duplicates):]

    def _find_sample_leaking_hash_impl(self):
        cols = [c for c in self.base_dataset.columns if c not in self.ignore_feats]

        base_dataset_hash = pd.util.hash_pandas_object(self.base_dataset[cols], index=False)
        base_dataset_hash = base_dataset_hash.drop_duplicates()

        test_dataset_hash = pd.util.hash_pandas_object(self.test_dataset[cols], index=False)
        test_dataset_hash = test_dataset_hash.drop_duplicates()

        concat_hashes = pd.concat([base_dataset_hash, test_dataset_hash])

        is_duplicated_concat = concat_hashes.duplicated(keep=False).values
        return is_duplicated_concat[len(base_dataset_hash):]

    def _verify_datasets(self):

        base_dataset_cols = set(self.base_dataset.columns)
        test_dataset_cols = set(self.test_dataset.columns)
        if not base_dataset_cols == set(self.test_dataset.columns):

            message_error = "Both datasets must contain the same columns. "
            base_cols_not_in_test = base_dataset_cols.difference(test_dataset_cols)
            if len(base_cols_not_in_test) > 0:
                message_error += f"Columns {base_cols_not_in_test} appear in base_dataset but not in test_dataset. "
            test_cols_not_in_base = test_dataset_cols.difference(base_dataset_cols)
            if len(test_cols_not_in_base) > 0:
                message_error += f"Columns {test_cols_not_in_base} appear in test_dataset but not in base_dataset. "

            raise ValueError(message_error)

    def _high_number_of_duplicates(self):

        if isinstance(self.threshold, int) or self.threshold.is_integer():
            return self._is_duplicated_test.sum() > self.threshold
        else:
            return (self._is_duplicated_test.sum() / len(self.test_dataset)) > self.threshold


class NoDuplicatesTest(RobustDataTest):
    """
    This test checks no duplicated samples are present in a dataframe. In case of
    having duplicated samples it could bias your model and/or evaluation metrics.

    Args:
        dataset (pandas.DataFrame): Dataset in which duplicates will be looked for.
        ignore_feats: List of columns that will be ignored when evaluating whether two samples are equal or not
        use_hash (bool): If True, it will create a hash for the rows in the dataframes in order to find duplicates. If False
            it will use `duplicate()` pandas method. Using hash usually results in faster execution in bigger datasets. If
            None it will use one method or other depending of the available memory
        name (str): A name for the test. If not used, it will take the name of the class.

    Example:
        ```python
        >>> from mercury.robust.data_tests import NoDuplicatesTest
        >>> test = NoDuplicatesTest(mydataframe)
        >>> test.run()
        >>> # Check test result details
        >>> test.info()
        ```
    """

    def __init__(
        self,
        dataset: "pandas.DataFrame",  # noqa: F821
        ignore_feats: List[str] = None,
        name: str = None,
        use_hash: bool = None,
        *args: Any, **kwargs: Any
    ):
        super().__init__(dataset, name, *args, **kwargs)
        self.ignore_feats = ignore_feats if ignore_feats else []
        if use_hash is None:
            use_hash = psutil.virtual_memory().percent > 50.0
        self.use_hash = use_hash
        self._num_duplicates = None

    def info(self):
        if self._num_duplicates is None:
            return None

        return {
            'num_duplicated': self._num_duplicates,
            'index_duplicates': self._index_duplicates
        }

    def run(self, *args, **kwargs):
        """
        Runs the test.

        Raises:
            FailedTestError if there exist duplicated samples in the provided dataframe
        """
        dataset = self.base_dataset[[x for x in self.base_dataset.columns if x not in self.ignore_feats]]
        if not self.use_hash:
            duplicates = dataset.duplicated()
        else:
            hashes = pd.util.hash_pandas_object(dataset, index=False)
            duplicates = hashes.duplicated()

        self._num_duplicates = duplicates.sum()
        self._index_duplicates = duplicates[duplicates].index

        if self._num_duplicates > 0:
            raise FailedTestError(
                f"Your dataset has {self._num_duplicates} duplicates. Drop or inspect them via the `info` method"
            )